# -*- coding: utf-8 -*-
# kobo-admin script, inspired by django-admin


import re
import os
import sys
import shutil

import kobo.cli
from kobo.types import Enum, EnumItem

import kobo.admin.commands


__all__ = (
    "TemplateError",
    "KoboAdminCommandContainer",
    "main",
    "copy_helper",
)


class TemplateError(Exception):
    """Processing of kobo directory structure template failed."""
    pass


# Inherit container to make sure nobody will change plugins I registered.
class KoboAdminCommandContainer(kobo.cli.CommandContainer):
    pass


def main():
    """Main method for kobo-admin script."""
    # Register plugins for commands
    KoboAdminCommandContainer.register_module(kobo.admin.commands, prefix="cmd_")

    command_container = KoboAdminCommandContainer()
    parser = kobo.cli.CommandOptionParser(
        command_container = command_container,
    )
    parser.run()
    return 0


def _camelize(name, fill_char=""):
    words = name.split('_')
    capwords = []
    for word in words:
        capwords.append(word.capitalize())
    return fill_char.join(capwords)


def _copy_file(path_old, path_new, name):
    # HACK: use .template suffix to prevent .py file byte compiling
    if path_new.endswith(".template"):
        path_new = path_new[:-9]

    fp_old = open(path_old, 'r')
    fp_new = open(path_new, 'w')

    # following django template sysntax
    fp_new.write(fp_old.read()\
        .replace('{{ project_name }}', name)\
        .replace('{{ project_name|upper }}', name.upper())\
        .replace('{{ project_name|camel }}', _camelize(name))\
        .replace('{{ project_name|camel_cmd }}', _camelize(name, fill_char="_"))
    )
    fp_old.close()
    fp_new.close()

    # copy permissions
    try:
        shutil.copymode(path_old, path_new)
    except OSError:
        sys.stderr.write("Notice: Couldn't set permission bits on %s. You're probably using an uncommon filesystem setup. No problem.\n" % path_new)


# based on django.core.management.base.copy_helper()
def copy_helper(name, directory, template_name):
    """Copies kobo based layout template into a specified directory."""

    # automatically convert dashes to underscores
    name = name.replace("-", "_")

    # name check from django
    if not re.search(r'^[_a-zA-Z]\w*$', name): # If it's not a valid directory name.
        # Provide a smart error message, depending on the error.
        if not re.search(r'^[_a-zA-Z]', name):
            message = 'make sure the name begins with a letter or underscore'
        else:
            message = 'use only numbers, letters and underscores'
        raise TemplateError("%r is not a valid name. Please %s." % (name, message))

    template_path = os.path.join(kobo.__path__[0], 'admin', 'templates', template_name)

    if os.path.isfile(template_path):
        path_new = os.path.join(directory, template_name).replace('__project_name__', name)
        path_new = path_new[path_new.find("@")+1:] # HACK: allow to have 2 source names (different prefix) with the same target names
        _copy_file(template_path, path_new, name)
        return

    top_dir = os.path.join(directory, name)
    try:
        os.mkdir(top_dir)
    except OSError as ex:
        raise TemplateError(ex)

    for dirname, subdirs, files in os.walk(template_path):
        relative_dir = dirname[len(template_path)+1:].replace('__project_name__', name)

        if relative_dir:
            os.mkdir(os.path.join(top_dir, relative_dir))

        for f in files:
            if f.endswith(".pyc") or f.endswith(".pyo"):
                continue
            path_old = os.path.join(dirname, f)
            path_new = os.path.join(top_dir, relative_dir, f).replace('__project_name__', name)
            _copy_file(path_old, path_new, name)



def call_if_callable(obj):
    """
    Determines if an object is callable, and returns its value or value of its call. 
    """
    if callable(obj):
        return obj()
    else:
        return obj



# -*- coding: utf-8 -*-


from __future__ import print_function
import base64
import six.moves.http_cookiejar as cookielib
import fcntl
import hashlib
import six.moves.http_client as httplib
import os
import socket
import ssl
import sys
import threading
import time
import six.moves.urllib.request as urllib2
import six.moves.xmlrpc_client as xmlrpclib
import six.moves.urllib.parse as urlparse

import kobo.shortcuts

try:
    import kerberos
    USE_KERBEROS = True
except ImportError:
    USE_KERBEROS = False


__all__ = (
    "CookieTransport",
    "SafeCookieTransport",
    "retry_request_decorator",
    "encode_xmlrpc_chunks_iterator",
    "decode_xmlrpc_chunk",
)


CONNECTION_LOCK = threading.Lock()


class TimeoutHTTPConnection(httplib.HTTPConnection):
    def connect(self):
        httplib.HTTPConnection.connect(self)
        timeout = getattr(self, "_timeout", 0)
        if timeout:
            self.sock.settimeout(timeout)

class TimeoutHTTPProxyConnection(TimeoutHTTPConnection):
    default_port = httplib.HTTPConnection.default_port

    def __init__(self, host, proxy, port=None, proxy_user=None, proxy_password=None, **kwargs):
        TimeoutHTTPConnection.__init__(self, proxy, **kwargs)
        self.proxy, self.proxy_port = self.host, self.port
        self.set_host_and_port(host, port)
        self.real_host, self.real_port = self.host, self.port
        self.proxy_user = proxy_user
        self.proxy_password = proxy_password

    def connect(self):
        # Connect to the proxy
        self.set_host_and_port(self.proxy, self.proxy_port)
        httplib.HTTPConnection.connect(self)
        self.set_host_and_port(self.real_host, self.real_port)
        timeout = getattr(self, "_timeout", 0)
        if timeout:
            self.sock.settimeout(timeout)

    def putrequest(self, method, url, skip_host=0, skip_accept_encoding=0):
        host = self.real_host
        if self.default_port != self.real_port:
            host = host + ':' + str(self.real_port)
        url = "http://%s%s" % (host, url)
        httplib.HTTPConnection.putrequest(self, method, url)
        self._add_auth_proxy_header()

    def _add_auth_proxy_header(self):
        if not self.proxy_user:
            return
        userpass = "%s:%s" % (self.proxy_user, self.proxy_password)
        encode_func = base64.encodebytes if hasattr(base64, "encodebytes") else base64.encodestring
        enc_userpass = encode_func(userpass).strip()
        self.putheader("Proxy-Authorization", "Basic %s" % enc_userpass)

    def set_host_and_port(self, host, port):
        """Due to httplib.py changes using set host & port method depends on package version"""
        if hasattr(self, "_set_hostport"):
            self._set_hostport(host, port)
        else:
            (self.host, self.port) = self._get_hostport(host, port)

class TimeoutHTTP(httplib.HTTPConnection):
    _connection_class = TimeoutHTTPConnection

    def set_timeout(self, timeout):
        self._conn._timeout = timeout


class TimeoutProxyHTTP(TimeoutHTTP):
    _connection_class = TimeoutHTTPProxyConnection

    def __init__(self, host='', proxy='',  port=None, strict=None,
                 proxy_user=None, proxy_password=None):
        if port == 0:
            port = None
        self._setup(self._connection_class(host, proxy, port=port,
                    strict=strict,
                    proxy_user=proxy_user,
                    proxy_password=proxy_password))

    def _setup(self, conn):
        httplib.HTTP._setup(self, conn)
        # XXX: Hack for python >= 2.7 where a _single_request method is used
        # and the method needs a connection object with .getresponse() method
        self.getresponse = conn.getresponse


class TimeoutHTTPSConnection(httplib.HTTPSConnection):
    def connect(self):
        httplib.HTTPSConnection.connect(self)
        timeout = getattr(self, "_timeout", 0)
        if timeout:
            self.sock.settimeout(timeout)


class TimeoutHTTPSProxyConnection(TimeoutHTTPProxyConnection):
    default_port = httplib.HTTPSConnection.default_port

    def __init__(self, host, proxy, port=None, proxy_user=None,
                 proxy_password=None, cert_file=None, key_file=None, **kwargs):
        TimeoutHTTPProxyConnection.__init__(self, host, proxy, port,
            proxy_user, proxy_password, **kwargs)
        self.cert_file = cert_file
        self.key_file = key_file
        self.connect()

    def connect(self):
        TimeoutHTTPProxyConnection.connect(self)
        host = "%s:%s" % (self.real_host, self.real_port)
        TimeoutHTTPConnection.putrequest(self, "CONNECT", host)
        self._add_auth_proxy_header()
        TimeoutHTTPConnection.endheaders(self)

        class MyHTTPSResponse(httplib.HTTPResponse):
            def begin(self):
                httplib.HTTPResponse.begin(self)
                self.will_close = 0

        response_class = self.response_class
        self.response_class = MyHTTPSResponse
        response = httplib.HTTPConnection.getresponse(self)
        self.response_class = response_class
        response.close()
        if response.status != 200:
            self.close()
            raise socket.error(1001, response.status, response.msg)

        self.sock = ssl.wrap_socket(self.sock, keyfile=self.key_file, certfile=self.cert_file)

    def putrequest(self, method, url, skip_host=0, skip_accept_encoding=0):
        return TimeoutHTTPConnection.putrequest(self, method, url)


class TimeoutHTTPS(httplib.HTTPSConnection):
    _connection_class = TimeoutHTTPSConnection

    def set_timeout(self, timeout):
        self._conn._timeout = timeout


class TimeoutProxyHTTPS(TimeoutHTTPS):
    _connection_class = TimeoutHTTPSProxyConnection

    def __init__(self, host='', proxy='',  port=None, strict=None,
                 proxy_user=None, proxy_password=None, cert_file=None,
                 key_file=None):
        if port == 0:
            port = None
        self._setup(self._connection_class(host, proxy, port=port,
                    strict=strict, proxy_user=proxy_user,
                    proxy_password=proxy_password, cert_file=cert_file,
                    key_file=key_file))

    def _setup(self, conn):
        httplib.HTTP._setup(self, conn)
        # XXX: Hack for python >= 2.7 where a _single_request method is used
        # and the method needs a connection object with .getresponse() method
        self.getresponse = conn.getresponse


class CookieResponse(object):
    """Fake response class for cookie extraction."""

    def __init__(self, headers):
        self.headers = headers

    def info(self):
        """Pass response headers to cookie jar."""
        return self.headers


class CookieTransport(xmlrpclib.Transport):
    """
    Cookie enabled XML-RPC transport.

    USAGE:
    >>> import xmlrpclib
        import kobo.xmlrpc
        client = xmlrpclib.ServerProxy("http://<server>/xmlrpc", transport=kobo.xmlrpc.CookieTransport())
        # for https:// connections use kobo.xmlrpc.SafeCookieTransport() instead.
    """

    _use_datetime = False # fix for python 2.5+
    scheme = "http"

    def __init__(self, *args, **kwargs):
        cookiejar = kwargs.pop("cookiejar", None)
        self.timeout = kwargs.pop("timeout", 0)
        self.proxy_config = self._get_proxy(**kwargs)
        self.no_proxy = os.environ.get("no_proxy", "").lower().split(',')
        self.context = kwargs.pop('context', None)

        self._cookie_headers = []
        self._krb_headers = []

        if hasattr(xmlrpclib.Transport, "__init__"):
            xmlrpclib.Transport.__init__(self, *args, **kwargs)

        self.cookiejar = cookiejar or cookielib.CookieJar()

        if hasattr(self.cookiejar, "load"):
            if not os.path.exists(self.cookiejar.filename):
                if hasattr(self.cookiejar, "save"):
                    self.cookiejar.save(self.cookiejar.filename)
            self.cookiejar.load(self.cookiejar.filename)

    def _get_proxy(self, **kwargs):
        """Return dict with appropriate proxy settings"""
        proxy = None
        proxy_user = None
        proxy_password = None

        if kwargs.get("proxy", None):
            # Use proxy from __init__ params
            proxy = kwargs["proxy"]
            if kwargs.get("proxy_user", None):
                proxy_user = kwargs["proxy_user"]
                if kwargs.get("proxy_password", None):
                    proxy_password = kwargs["proxy_user"]
        else:
            # Try to get proxy settings from environmental vars
            if self.scheme == "http" and os.environ.get("http_proxy", None):
                proxy = os.environ["http_proxy"]
            elif self.scheme == "https" and os.environ.get("https_proxy", None):
                proxy = os.environ["https_proxy"]

        if proxy:
            # Parse proxy address
            # e.g. http://user:password@proxy.company.com:8001/foo/bar

            # Get raw location without path
            location = urlparse.urlparse(proxy)[1]
            if not location:
                # proxy probably doesn't have a protocol in prefix
                location = urlparse.urlparse("http://%s" % proxy)[1]

            # Parse out username and password if present
            if '@' in location:
                userpas, location = location.split('@', 1)
                if userpas and location and not proxy_user:
                    # Set proxy user only if proxy_user is not set yet
                    proxy_user = userpas
                    if ':' in userpas:
                        proxy_user, proxy_password = userpas.split(':', 1)

            proxy = location

        proxy_settings = {
            "proxy": proxy,
            "proxy_user": proxy_user,
            "proxy_password": proxy_password,
        }

        return proxy_settings

    def make_connection(self, host):
        host.lower()
        host_ = host  # Host with(out) port
        if ':' in host:
            # Remove port from the host
            host_ = host.split(':')[0]
        else:
            host_ = "%s:%s" % (host, TimeoutHTTPProxyConnection.default_port)

        if self.proxy_config["proxy"] and host not in self.no_proxy and host_ not in self.no_proxy:
            if sys.version_info[:2] < (2, 7):
                host, extra_headers, x509 = self.get_host_info(host)
                conn = TimeoutProxyHTTP(host, **self.proxy_config)
                conn.set_timeout(self.timeout)
                return conn
            else:
                CONNECTION_LOCK.acquire()
                host, extra_headers, x509 = self.get_host_info(host)
                conn = TimeoutProxyHTTPS(host, **self.proxy_config)
                conn.set_timeout(self.timeout)
                CONNECTION_LOCK.release()
                return conn

        if sys.version_info[:2] < (2, 7):
            host, extra_headers, x509 = self.get_host_info(host)
            conn = TimeoutHTTP(host)
            conn.set_timeout(self.timeout)
            return conn
        else:
            CONNECTION_LOCK.acquire()
            self._connection = (None, None) # this disables connection caching which causes a race condition when running in threads
            conn = xmlrpclib.Transport.make_connection(self, host)
            CONNECTION_LOCK.release()
            if self.timeout:
                conn.timeout = self.timeout
            return conn

    def send_cookies(self, connection, cookie_request):
        """Add cookies to the header."""
        self.cookiejar.add_cookie_header(cookie_request)

        for header, value in cookie_request.header_items():
            if header.startswith("Cookie"):
                connection.putheader(header, value)

    def _save_cookies(self, headers, cookie_request):
        cookie_response = CookieResponse(headers)
        self.cookiejar.extract_cookies(cookie_response, cookie_request)
        if hasattr(self.cookiejar, "save"):
            self.cookiejar.save(self.cookiejar.filename)

    def _kerberos_client_request(self, host, handler, errcode, errmsg, headers):
        """Kerberos auth - create a client request string"""

        # check if "Negotiate" challenge is present in headers
        negotiate = [i.lower() for i in headers.get("WWW-Authenticate", "").split(", ")]
        if "negotiate" not in negotiate:
            # negotiate not supported, raise 401 error
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

        # initialize GSSAPI
        service = "HTTP@%s" % host
        rc, vc = kerberos.authGSSClientInit(service)
        if rc != 1:
            errmsg = "KERBEROS: Could not initialize GSSAPI"
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

        # do a client step
        rc = kerberos.authGSSClientStep(vc, "")
        if rc != 0:
            errmsg = "KERBEROS: Client step failed"
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

        return vc, kerberos.authGSSClientResponse(vc)

    def _kerberos_verify_response(self, vc, host, handler, errcode, errmsg, headers):
        """Kerberos auth - verify client identity"""
        # verify that headers contain WWW-Authenticate header
        auth_header = headers.get("WWW-Authenticate", None)
        if auth_header is None:
            errcode = 401
            errmsg = "KERBEROS: No WWW-Authenticate header in second HTTP response"
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

        # verify that WWW-Authenticate contains Negotiate
        splits = auth_header.split(" ", 1)
        if (len(splits) != 2) or (splits[0].lower() != "negotiate"):
            errcode = 401
            errmsg = "KERBEROS: Incorrect WWW-Authenticate header in second HTTP response: %s" % auth_header
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

        # do another client step to verify response from server
        errmsg = "KERBEROS: Could not verify server WWW-Authenticate header in second HTTP response"
        try:
            rc = kerberos.authGSSClientStep(vc, splits[1])
            if rc == -1:
                errcode = 401
                raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)
        except kerberos.GSSError as ex:
            errcode = 401
            errmsg += ": %s/%s" % (ex[0][0], ex[1][0])
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

        # cleanup
        rc = kerberos.authGSSClientClean(vc)
        if rc != 1:
            errcode = 401
            errmsg = "KERBEROS: Could not clean-up GSSAPI: %s/%s" % (ex[0][0], ex[1][0])
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

    def _request(self, host, handler, request_body, verbose=0):
        """Send a HTTP request."""
        h = self.make_connection(host)
        self.verbose = verbose
        if verbose:
            h.set_debuglevel(1)

        request_url = "%s://%s/" % (self.scheme, host)
        cookie_request = urllib2.Request(request_url)

        self.send_request(h, handler, request_body)
        self.send_host(h, host)
        self.send_cookies(h, cookie_request)
        self.send_user_agent(h)
        self.send_content(h, request_body)
        errcode, errmsg, headers = h.getreply()

        if errcode == 401 and USE_KERBEROS:
            vc, challenge = self._kerberos_client_request(host, handler, errcode, errmsg, headers)
            # retry the original request & add the Authorization header:
            self.send_request(h, handler, request_body)
            self.send_host(h, host)
            h.putheader("Authorization", "Negotiate %s" % challenge)
            self.send_cookies(h, cookie_request)
            self.send_user_agent(h)
            self.send_content(h, request_body)
            errcode, errmsg, headers = h.getreply()
            self._kerberos_verify_response(vc, host, handler, errcode, errmsg, headers)

        elif errcode != 200:
            raise xmlrpclib.ProtocolError(host + handler, errcode, errmsg, headers)

        self._save_cookies(headers, cookie_request)

        try:
            sock = h._conn.sock
        except AttributeError:
            sock = None

        return self._parse_response(h.getfile(), sock)

    def _single_request(self, host, handler, request_body, verbose=0):
        # issue XML-RPC request

        request_url = "%s://%s/" % (self.scheme, host)
        cookie_request = urllib2.Request(request_url)

        h = self.make_connection(host)
        self.verbose = verbose
        if verbose:
            h.set_debuglevel(1)

        try:
            self.send_request(h, handler, request_body)
            self.send_host(h, host)
            self.send_cookies(h, cookie_request)
            self.send_user_agent(h)
            self.send_content(h, request_body)

            response = h.getresponse(buffering=True)

            if response.status == 401 and USE_KERBEROS:
                vc, challenge = self._kerberos_client_request(host, handler, response.status, response.reason, response.msg)

                # discard any response data
                if (response.getheader("content-length", 0)):
                    response.read()

                # retry the original request & add the Authorization header:
                self.send_request(h, handler, request_body)
                self._extra_headers = [("Authorization", "Negotiate %s" % challenge)]
                self.send_host(h, host)
                self.send_user_agent(h)
                self.send_content(h, request_body)
                self._extra_headers = []
                response = h.getresponse(buffering=True)
                self._kerberos_verify_response(vc, host, handler, response.status, response.reason, response.msg)

            if response.status == 200:
                self.verbose = verbose
                self._save_cookies(response.msg, cookie_request)
                return self.parse_response(response)
        except xmlrpclib.Fault:
            raise
        except Exception:
            # All unexpected errors leave connection in
            # a strange state, so we clear it.
            self.close()
            raise

        # discard any response data and raise exception
        if (response.getheader("content-length", 0)):
            response.read()
        raise xmlrpclib.ProtocolError(host + handler, response.status, response.reason, response.msg)

    def _single_request3(self, host, handler, request_body, verbose=False):
        # issue XML-RPC request

        request_url = "%s://%s/" % (self.scheme, host)
        cookie_request = urllib2.Request(request_url)

        self.prepare_cookies(cookie_request)

        self.verbose = verbose

        try:
            h = self.send_request(host, handler, request_body, verbose)

            response = h.getresponse()

            if response.status == 401 and USE_KERBEROS:
                vc, challenge = self._kerberos_client_request(host, handler, response.status, response.reason, response.msg)

                # discard any response data
                if (response.getheader("content-length", 0)):
                    response.read()

                # retry the original request & add the Authorization header:
                self._krb_headers = [("Authorization", "Negotiate %s" % challenge)]
                h = self.send_request(host, handler, request_body, verbose)
                self._krb_headers = []

                response = h.getresponse()
                self._kerberos_verify_response(vc, host, handler, response.status, response.reason, response.msg)

            if response.status == 200:
                self.verbose = verbose
                self._save_cookies(response.msg, cookie_request)
                return self.parse_response(response)
        except xmlrpclib.Fault:
            raise
        except Exception:
            # All unexpected errors leave connection in
            # a strange state, so we clear it.
            self.close()
            raise

        # discard any response data and raise exception
        if (response.getheader("content-length", 0)):
            response.read()
        raise xmlrpclib.ProtocolError(host + handler, response.status, response.reason, response.msg)

    # override the appropriate request method
    if sys.version_info[0] >= 3:
        single_request = _single_request3
    elif hasattr(xmlrpclib.Transport, "single_request"):
        # python 2.7+
        single_request = _single_request
    else:
        # python 2.6-
        request = _request

    def send_headers(self, connection, headers):
        headers.extend(self._cookie_headers)
        headers.extend(self._krb_headers)
        super().send_headers(connection, headers)

    def prepare_cookies(self, cookie_request):
        """Add cookies to the header."""
        self.cookiejar.add_cookie_header(cookie_request)

        self._cookie_headers = []

        for header, value in cookie_request.header_items():
            if header.startswith("Cookie"):
                self._cookie_headers.append((header, value))


class SafeCookieTransport(CookieTransport, xmlrpclib.SafeTransport):
    """
    Cookie enabled XML-RPC transport over HTTPS.

    USAGE: see CookieTransport
    """
    scheme = "https"

    def make_connection(self, host):
        host.lower()
        host_ = host  # Host with(out) port
        if ':' in host:
            # Remove port from the host
            host_ = host.split(':')[0]
        else:
            host_ = "%s:%s" % (host, TimeoutHTTPSProxyConnection.default_port)

        if self.proxy_config["proxy"] and host not in self.no_proxy and host_ not in self.no_proxy:
            if sys.version_info[:2] < (2, 7):
                host, extra_headers, x509 = self.get_host_info(host)
                conn = TimeoutProxyHTTPS(host, **self.proxy_config)
                conn.set_timeout(self.timeout)
                return conn
            else:
                CONNECTION_LOCK.acquire()
                host, extra_headers, x509 = self.get_host_info(host)
                conn = TimeoutProxyHTTPS(host, **self.proxy_config)
                conn.set_timeout(self.timeout)
                CONNECTION_LOCK.release()
                return conn

        if sys.version_info[:2] < (2, 7):
            host, extra_headers, x509 = self.get_host_info(host)
            conn = TimeoutHTTPS(host, None, **(x509 or {}))
            conn.set_timeout(self.timeout)
            return conn
        else:
            CONNECTION_LOCK.acquire()
            self._connection = (None, None) # this disables connection caching which causes a race condition when running in threads
            conn = xmlrpclib.SafeTransport.make_connection(self, host)
            if self.timeout:
                conn.timeout = self.timeout
            CONNECTION_LOCK.release()
            return conn

    def __init__(self, *args, **kwargs):
        self.context = kwargs.pop('context', None)
        xmlrpclib.SafeTransport.__init__(self, *args, **kwargs)
        CookieTransport.__init__(self, *args, **kwargs)


def retry_request_decorator(transport_class):
    """Use this class decorator on a Transport to retry requests which failed on socket errors."""
    class RetryTransportClass(transport_class):
        def __init__(self, *args, **kwargs):
            self.retry_count = kwargs.pop("retry_count", 5)
            self.retry_timeout = kwargs.pop("retry_timeout", 30)
            if hasattr(transport_class, "__init__"):
                transport_class.__init__(self, *args, **kwargs)

        def request(self, *args, **kwargs):
            if self.retry_count == 0:
                return transport_class.request(self, *args, **kwargs)

            for i in range(self.retry_count + 1):
                try:
                    result = transport_class.request(self, *args, **kwargs)
                    return result
                except KeyboardInterrupt:
                    raise
                except (socket.error, socket.herror, socket.gaierror, socket.timeout,
                        httplib.CannotSendRequest, xmlrpclib.ProtocolError) as ex:
                    if i >= self.retry_count:
                        raise
                    retries_left = self.retry_count - i
                    retries = "%d %s left" % (retries_left, retries_left == 1 and "retry" or "retries") # 1 retry left / X retries left
                    print("XML-RPC connection to %s failed: %s, %s" % (args[0], ex.args[1:], retries), file=sys.stderr)
                    time.sleep(self.retry_timeout)

    RetryTransportClass.__name__ = transport_class.__name__
    RetryTransportClass.__doc__ = transport_class.__name__
    return RetryTransportClass


def encode_xmlrpc_chunks_iterator(file_obj):
    """
    Prepare data for a xml-rpc transfer.
    Iterate through (chunk_start, chunk_len, chunk_checksum, encoded_chunk) tuples.
    Final tuple is (total_length, -1, total_checksum, "").

    @param file_obj: file object (or StringIO, etc.)
    @type  file_obj: file
    @return: (chunk_start, chunk_len, chunk_checksum, encoded_chunk)
    @rtype:  (str, str, str, str)
    """

    CHUNK_SIZE = 1024 ** 2
    checksum = hashlib.sha256()
    chunk_start = file_obj.tell()

    while True:
        chunk = file_obj.read(CHUNK_SIZE)
        if not chunk:
            break
        checksum.update(chunk)
        encode_func = base64.encodebytes if hasattr(base64, "encodebytes") else base64.encodestring
        encoded_chunk = encode_func(chunk)
        yield (str(chunk_start), str(len(chunk)), hashlib.sha256(chunk).hexdigest().lower(), encoded_chunk)
        chunk_start += len(chunk)

    yield (str(chunk_start), -1, checksum.hexdigest().lower(), "")


def decode_xmlrpc_chunk(chunk_start, chunk_len, chunk_checksum, encoded_chunk, write_to=None, mode=0o644):
    """
    Decode a data chunk and optionally write it to a file.

    @param chunk_start: chunk start position in the file (-1 for append)
    @type  chunk_start: str
    @param chunk_len: chunk length
    @type  chunk_len: str
    @param chunk_checksum: sha256 checksum (lower case)
    @type  chunk_checksum: str
    @param encoded_chunk: base64 encoded chunk
    @type  encoded_chunk: str
    @param write_to: path to a file in which the decoded data will be written
    @type  write_to: str
    @param mode: file permissions (example: 0644)
    @type  mode: int
    @return: decoded data
    @rtype:  str
    """

    chunk_start = int(chunk_start)
    chunk_len = int(chunk_len)
    decode_func = base64.decodebytes if hasattr(base64, "decodebytes") else base64.decodestring

    if isinstance(encoded_chunk, xmlrpclib.Binary):
        chunk = decode_func(encoded_chunk.data)
    else:
        chunk = decode_func(encoded_chunk)

    if chunk_len not in (-1, len(chunk)):
        raise ValueError("Chunk length doesn't match.")

    if chunk_len == -1:
        chunk = ""
    elif chunk_checksum != hashlib.sha256(chunk).hexdigest().lower():
        raise ValueError("Chunk checksum doesn't match.")

    if not write_to:
        return chunk

    # code below handles writing to a file

    target_dir = os.path.dirname(write_to)
    if not os.path.isdir(target_dir):
        try:
            os.makedirs(target_dir, mode=0o755)
        except OSError as ex:
            if ex.errno != 17:
                raise

    fd = os.open(write_to, os.O_RDWR | os.O_CREAT, mode)
    fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    try:
        if chunk_start != -1:
            os.ftruncate(fd, chunk_start)
        os.lseek(fd, 0, 2) # 2=os.SEEK_END
        os.write(fd, chunk)
    finally:
        fcntl.lockf(fd, fcntl.LOCK_UN)
        os.close(fd)

    if chunk_start != -1 and chunk_len == -1:
        # final chunk, compute checksum of whole file
        file_checksum = kobo.shortcuts.compute_file_checksums(write_to, ["sha256"])["sha256"]
        if file_checksum != chunk_checksum:
            raise ValueError("File checksum does not match.")

    return chunk



import gc
import itertools
import json
import locale
import os

from tempfile import TemporaryFile
from gzip import GzipFile
from io import BytesIO
from shutil import rmtree

import mock
import six

import django.conf
import django.test

from django.http import HttpResponse

from django.contrib.auth.models import User

from kobo.hub.models import Task, Arch, Channel, TASK_STATES

from .utils import DjangoRunner, profile

runner = DjangoRunner()
setup_module = runner.start
teardown_module = runner.stop

TASK_ID = 123


def tiny_log_content():
    """Returns very small log content.  Particularly, small enough that
    gzip-compressing will *increase* the content size."""
    return 'hi!'


def small_log_content():
    """Returns content for a small log, where small means able to fit into
    a single chunk (of whatever chunking scheme might be in use)."""
    return '\n'.join([
        'Line 1 of test log',
        'Line 2 of test log',
    ])


def big_log_content():
    """Returns content for a big log likely to span multiple chunks."""
    out = 'some log content'
    i = 1
    while len(out) < (1024**2) * 20:
        out = ('\njoin %d\n' % i).join([out, out])
        i = i + 1
    return ''.join(['BEGIN LOG\n', out, 'END LOG\n'])


def html_content():
    """Content suitable for an HTML formatted log."""
    return '<html>This is some HTML content</html>'


def response_content(response):
    """Get all the content from an HttpResponse (eagerly loaded if the
    response is streaming)."""
    if response.streaming:
        content = b''.join(response.streaming_content)
    else:
        content = response.content

    if six.PY3:
        try:
            return str(content, encoding=response.charset)
        except UnicodeDecodeError:
            return content
    else:
        return content


def gzip_decompress(data):
    """Like zlib.decompress but handles the gzip header, so it can be used
    on gzipped log files."""
    io = BytesIO(data)
    gzfile = GzipFile(fileobj=io)
    return gzfile.read()


class TestViewLog(django.test.TransactionTestCase):
    def __init__(self, *args, **kwargs):
        super(TestViewLog, self).__init__(*args, **kwargs)
        # This is cached since it is a bit slow to build
        self.big_log_content = big_log_content()

    def cleanDataDirs(self):
        """Delete the log directory for the test task."""
        log_dir = os.path.join(django.conf.settings.TASK_DIR, '0', '0', str(TASK_ID))
        if os.path.exists(log_dir):
            rmtree(log_dir)

    def tearDown(self):
        self.cleanDataDirs()

        super(TestViewLog, self).tearDown()

    def setUp(self):
        self._fixture_teardown()
        super(TestViewLog, self).setUp()

        self.cleanDataDirs()

        # These objects are required to exist but their attributes are irrelevant
        user = User.objects.create()
        arch = Arch.objects.create(name='testarch')
        channel = Channel.objects.create(name='testchannel')

        test_task = Task.objects.create(
            pk=TASK_ID, arch=arch, channel=channel, owner=user)

        # associate some compressed and decompressed logs with the task.
        # (the save and gzip_logs methods here are writing files to disk)
        test_task.logs['zipped_tiny.log'] = tiny_log_content()
        test_task.logs['zipped_small.log'] = small_log_content()
        test_task.logs['zipped_big.log'] = self.big_log_content
        test_task.logs.save()
        test_task.logs.gzip_logs()

        test_task.logs['small.log'] = small_log_content()
        test_task.logs['big.log'] = self.big_log_content
        test_task.logs['log.html'] = html_content()
        test_task.logs.save()

        self.client = django.test.Client()

        # for more accurate memory_profiler tests
        gc.collect()

    def test_view_zipped_small_raw(self):
        """Fetching a small compressed log with raw format should yield
        the gzip-compressed content."""
        response, content = self.assertGetLog(
            'zipped_small.log',
            data={'format': 'raw'},
        )

        uncompressed_content = gzip_decompress(content)
        if six.PY3:
            uncompressed_content = str(uncompressed_content,
                                       encoding=locale.getpreferredencoding())
        self.assertEqual(uncompressed_content, small_log_content())

    def test_view_small_raw_offset(self):
        """Fetching a small log with an offset should return only the
        content from that offset to the end of the log."""
        full_content = small_log_content()

        offset = 10

        self.assertGetLog(
            'small.log',
            data={'format': 'raw', 'offset': offset},
            expected_content=full_content[offset:]
        )

    def test_view_html_passthrough(self):
        """Fetching an HTML log yields exactly the HTML content,
        not wrapped in any template (even if format: raw is not requested)."""
        self.assertGetLog(
            'log.html',
            expected_content=html_content()
        )

    def test_view_html_wrapped(self):
        """In the typical case of fetching a small plaintext log with no
        specified format, the log's text is wrapped in an HTML view."""
        response, content = self.assertGetLog(
            'small.log',
            # Django-rendered template responses are not including
            # Content-Length
            test_content_length=False,
        )

        # The response is expected to be wrapped in some view.
        # We are only doing a very basic verification of the view here
        self.assertTrue(content.startswith('<!DOCTYPE html'))
        self.assertTrue((small_log_content() + "\n</pre>") in content, content)

        # No trimming necessary
        self.assertFalse('...trimmed, download required for full log' in content)

    def test_view_tiny_html_wrapped(self):
        """As typical case, but with a log file so small that the gzipped file
        is larger than the uncompressed content."""
        response, content = self.assertGetLog(
            'zipped_tiny.log',
            test_content_length=False,
        )

        self.assertTrue(content.startswith('<!DOCTYPE html'))
        self.assertTrue((tiny_log_content() + "\n</pre>") in content, content)
        self.assertFalse('...trimmed, download required for full log' in content)

    @profile
    def test_view_big_html_wrapped(self):
        response = self.get_log('big.log')
        self.assertEqual(response.status_code, 200, response)

        content = response.content
        if six.PY3:
            content = str(content, encoding=response.charset)
        self.assertTrue(content.startswith('<!DOCTYPE html'))

        # Should not have returned whole content
        self.assertTrue(len(content) < len(self.big_log_content))

        # Should tell the user that the response was trimmed
        self.assertTrue('&lt;...trimmed, download required for full log&gt;\n' in content)

        # Content should contain the end of the log file
        self.assertTrue(self.big_log_content[-5000:] in content)

        # Content should NOT contain the beginning of the log file
        self.assertFalse(self.big_log_content[0:5000] in content)

    @profile
    def test_view_big_html_wrapped_with_offset(self):
        # Just get the last 2000 chars - this should not trigger log trimming
        offset = len(big_log_content()) - 2000
        response = self.get_log('big.log', data={'offset': offset})

        content = response.content
        if six.PY3:
            content = str(content, encoding=response.charset)
        self.assertTrue(content.startswith('<!DOCTYPE html'))
        self.assertTrue(big_log_content()[-2000:] in content)

    @profile
    def test_view_zipped_big_html_context(self):
        """Verify the context passed into HTML template contains correct values."""
        def render(*args, **kwargs):
            return HttpResponse(status=200)

        with mock.patch('kobo.hub.views.render', side_effect=render) as render_mock:
            self.get_log('zipped_big.log')

        mock_call = render_mock.mock_calls[0]

        # make sure we're looking at the right call
        self.assertEqual(mock_call[0], '')

        call_args = mock_call[1]
        (_, template_name, context) = call_args

        self.assertEqual(template_name, 'task/log.html')

        # Check various things passed in context

        # The task is not yet completed
        self.assertEqual(context['task_finished'], 0)

        # It should ask log watcher to poll soon
        self.assertEqual(context['next_poll'], 5000)

        # Since it provided the end of the log content, it should give an offset
        # pointing just past the end
        self.assertEqual(context['offset'], len(self.big_log_content))

    @profile
    def test_view_big_raw(self):
        big_content = self.big_log_content
        offset = 0
        response = self.get_log('big.log', data={'format': 'raw'})
        self.assertEqual(response.status_code, 200, response)

        # Response should be streamed in chunks
        self.assertTrue(response.streaming)
        count = 0
        for chunk in response.streaming_content:
            chunklen = len(chunk)
            if six.PY3:
                chunk = str(chunk, encoding=locale.getpreferredencoding())
            self.assertEqual(chunk, big_content[offset:offset + chunklen])
            offset = offset + chunklen
            count = count + 1

        self.assertTrue(count > 1)

    @profile
    def test_view_zipped_big_raw(self):
        big_content = self.big_log_content

        response = self.get_log('zipped_big.log', data={'format': 'raw'})
        self.assertEqual(response.status_code, 200, response)

        # Response should be streamed in chunks, and once decompressed,
        # should be the expected content
        self.assertTrue(response.streaming)

        tempfile = TemporaryFile()

        for chunk in response.streaming_content:
            tempfile.write(chunk)

        tempfile.seek(0)
        gz_file = GzipFile(mode='r', fileobj=tempfile)
        all_data = gz_file.read()
        if six.PY3:
            all_data = str(all_data, encoding=locale.getpreferredencoding())
        self.assertEqual(all_data, big_content)

    def test_view_zipped_small_json(self):
        """Fetching compressed small log via JSON gives expected document."""
        response, content = self.assertGetLog(
            'zipped_small.log',
            view_type='log-json',
            test_content_length=False,
        )
        expected_content = small_log_content()
        doc = json.loads(content)
        self.assertEqual(doc, {
            'content': expected_content,
            'new_offset': len(expected_content),
            'task_finished': 0,
            'next_poll': 5000,
        })

    def test_view_small_json(self):
        """As test_view_zipped_small_json, but log is uncompressed.
        Otherwise identical, i.e. compression is expected to be handled
        transparently."""
        response, content = self.assertGetLog(
            'small.log',
            view_type='log-json',
            test_content_length=False,
        )
        expected_content = small_log_content()
        doc = json.loads(content)
        self.assertEqual(doc, {
            'content': expected_content,
            'new_offset': len(expected_content),
            'task_finished': 0,
            'next_poll': 5000,
        })

    def test_view_small_json_offset(self):
        """Fetching log with offset via JSON returns only the content
        starting from that offset."""
        full_content = small_log_content()
        offset = 10
        expected_content = full_content[offset:]

        response, content = self.assertGetLog(
            'small.log',
            view_type='log-json',
            test_content_length=False,
            data={'offset': offset}
        )

        doc = json.loads(content)
        self.assertEqual(doc, {
            'content': expected_content,
            'new_offset': offset + len(expected_content),
            'task_finished': 0,
            'next_poll': 5000,
        })

    @profile
    def test_view_big_json(self):
        response, content = self.assertGetLog(
            'big.log',
            view_type='log-json',
            test_content_length=False,
        )
        doc = json.loads(content)

        # Should have given us a subset of the content and told us to poll
        # again soon
        self.assertTrue(doc['content'] in self.big_log_content)
        self.assertEqual(doc['new_offset'], len(doc['content'])),
        self.assertEqual(doc['task_finished'], 0)
        self.assertEqual(doc['next_poll'], 0)

    @profile
    def test_view_zipped_big_json(self):
        response, content = self.assertGetLog(
            'zipped_big.log',
            view_type='log-json',
            test_content_length=False,
        )
        doc = json.loads(content)

        self.assertTrue(doc['content'] in self.big_log_content)
        self.assertEqual(doc['new_offset'], len(doc['content'])),
        self.assertEqual(doc['task_finished'], 0)
        self.assertEqual(doc['next_poll'], 0)

    @profile
    def test_view_zipped_big_json_offset(self):
        offset = 20000
        response, content = self.assertGetLog(
            'zipped_big.log',
            view_type='log-json',
            test_content_length=False,
            data={'offset': offset}
        )
        doc = json.loads(content)

        # We're not verifying exactly how much we expect to be given, but we do expect it's
        # a subset of the available content
        content_len = len(doc['content'])
        self.assertTrue(content_len < len(self.big_log_content) - offset)
        self.assertEqual(doc['content'], self.big_log_content[offset:offset + content_len])
        self.assertEqual(doc['new_offset'], offset + content_len),
        self.assertEqual(doc['task_finished'], 0)
        self.assertEqual(doc['next_poll'], 0)

    @profile
    def test_view_big_json_iterate(self):
        """Iterates over log chunks based on next_poll and new_offset in responses,
        simulating the LogWatcher JS behavior. Verifies that the full log can be assembled."""

        # The task has to be finished for this test, otherwise it'll loop forever
        task = Task.objects.get(id=TASK_ID)
        task.state = TASK_STATES["CLOSED"]
        task.save()

        all_content = ''
        offset = 0

        for i in itertools.count():
            self.assertTrue(i < 10000, 'infinite loop reading log?')

            response, content = self.assertGetLog(
                'big.log',
                view_type='log-json',
                test_content_length=False,
                data={'offset': offset},
            )
            doc = json.loads(content)
            all_content = all_content + doc['content']
            offset = doc['new_offset']
            if doc['next_poll'] is None:
                break

        self.assertEqual(all_content, self.big_log_content)

    def assertGetLog(self, log_name, view_type='log', test_content_length=True,
                     expected_content=None, data={}):
        """Verify log can be successfully retrieved and response has certain properties.
        The HttpResponse and its content are returned so that tests may do further
        assertions."""
        response = self.get_log(log_name, view_type=view_type, data=data)
        content = response_content(response)

        # Should always succeed
        self.assertEqual(response.status_code, 200)

        if expected_content:
            # Should be exactly this content
            # (if not provided, caller is expected to do custom verification)
            self.assertEqual(content, expected_content)

        # Content-Length is optional, but if present, must be correct
        if test_content_length or ('Content-Length' in response):
            self.assertEqual(response['Content-Length'], str(len(content)))

        return response, content

    def get_log(self, log_name, view_type='log', data={}):
        url = '/task/{0}/{1}/{2}'.format(TASK_ID, view_type, log_name)
        return self.client.get(url, data)



import pytest

from kobo.django.forms import JSONWidget


@pytest.mark.parametrize(("value", "output"), [
    ('{"a": "b"}', '<textarea id="noid" name="test_widget">{&quot;a&quot;: &quot;b&quot;}</textarea>'),
    ({"a": "b"}, '<textarea id="noid" name="test_widget">{&quot;a&quot;: &quot;b&quot;}</textarea>'),
    ("[1, 2, 3]", '<textarea id="noid" name="test_widget">[1, 2, 3]</textarea>'),
    ([1, 2, 3], '<textarea id="noid" name="test_widget">[1, 2, 3]</textarea>'),
])
def test_JSONWidget(value, output):
    w = JSONWidget()
    assert w.render("test_widget", value, attrs={"id": "noid"}) == output



# -*- coding: utf-8 -*-

import django

from django.core.exceptions import PermissionDenied
from django.contrib.auth.models import User

from mock import PropertyMock

from kobo.client.constants import TASK_STATES
from kobo.hub.models import Arch, Channel, Task, Worker
from kobo.hub.xmlrpc import client

from .utils import DjangoRunner

runner = DjangoRunner()
setup_module = runner.start
teardown_module = runner.stop


def _make_request(user=None, is_authenticated=True, is_superuser=True, meta=None):
    if user is None:
        return PropertyMock(
            user=PropertyMock(
                is_authenticated=lambda: is_authenticated,
                is_superuser=is_superuser,
                username='testuser',
            ),
            META=meta,
        )

    return PropertyMock(
        user=user,
        META=meta,
    )


class TestXmlRpcClient(django.test.TransactionTestCase):

    def setUp(self):
        self._fixture_teardown()
        super(TestXmlRpcClient, self).setUp()

        user = User.objects.create(username='testuser', is_superuser=True)
        arch = Arch.objects.create(name='noarch', pretty_name='noarch')
        channel = Channel.objects.create(name='default')

        self._arch = arch
        self._channel = channel
        self._user = user

    def test_shutdown_worker(self):
        w = Worker.objects.create(
            worker_key='worker',
            name='worker',
            enabled=True,
        )

        task_id = client.shutdown_worker(_make_request(), w.name)
        self.assertTrue(task_id > 0)

        task = Task.objects.get(id=task_id)
        self.assertTrue(task.exclusive)
        self.assertTrue('kill' in task.args)
        self.assertFalse(task.args['kill'])
        self.assertEqual(task.method, 'ShutdownWorker')

    def test_shutdown_worker_kill(self):
        w = Worker.objects.create(
            worker_key='enabled-worker',
            name='enabled-worker',
            enabled=True,
        )

        task_id = client.shutdown_worker(_make_request(), w.name, kill=True)
        self.assertTrue(task_id > 0)

        task = Task.objects.get(id=task_id)
        self.assertTrue(task.exclusive)
        self.assertTrue('kill' in task.args)
        self.assertTrue(task.args['kill'])
        self.assertEqual(task.method, 'ShutdownWorker')

    def test_enable_worker(self):
        w = Worker.objects.create(
            worker_key='disabled-worker',
            name='disabled-worker',
            enabled=False,
        )
        self.assertFalse(w.enabled)

        client.enable_worker(_make_request(), w.name)

        w = Worker.objects.get(id=w.id)
        self.assertTrue(w.enabled)

    def test_disable_worker(self):
        w = Worker.objects.create(
            worker_key='enabled-worker',
            name='enabled-worker',
            enabled=True,
        )
        self.assertTrue(w.enabled)

        client.disable_worker(_make_request(), w.name)

        w = Worker.objects.get(id=w.id)
        self.assertFalse(w.enabled)

    def test_get_worker_info(self):
        w = Worker.objects.create(
            worker_key='test-worker',
            name='test-worker',
        )

        info = client.get_worker_info(_make_request(), w.name)
        self.assertEqual(info['id'], w.id)

    def test_get_worker_info_non_existent(self):
        info = client.get_worker_info(_make_request(), 'non-existent')
        self.assertEqual(info, {})

    def test_get_worker_info_no_auth(self):
        # Should be permissible if anonymous - no exception raised
        client.get_worker_info(_make_request(
            is_authenticated=False,
            is_superuser=False,
        ), 'worker')

    def test_get_worker_info_if_authenticated_but_not_admin(self):
        # Should be permissible if authenticated, but not admin - no exception raised
        client.get_worker_info(_make_request(
            is_authenticated=False,
            is_superuser=False,
        ), 'worker')

    def test_task_info(self):
        task_id = Task.create_task(self._user.username, 'label', 'method')
        task_info = client.task_info(_make_request(), task_id)
        self.assertEqual(task_info['id'], task_id)

    def test_task_info_flat(self):
        task_id = Task.create_task(self._user.username, 'label', 'method')
        task_info = client.task_info(_make_request(), task_id, flat=True)
        self.assertEqual(task_info['id'], task_id)

    def test_task_info_non_existent(self):
        with self.assertRaises(Task.DoesNotExist):
            client.task_info(_make_request(), 999)

    def test_get_tasks_with_no_filters(self):
        Task.create_task(self._user.username, 'task-1', 'method')
        Task.create_task(self._user.username, 'task-2', 'method')
        Task.create_task(self._user.username, 'task-3', 'method')
        Task.create_task(self._user.username, 'task-4', 'method')
        Task.create_task(self._user.username, 'task-5', 'method')

        task_list = client.get_tasks(_make_request(), None, None)

        self.assertEqual(len(task_list), 5)

    def test_get_tasks_filter_by_ids(self):
        t1 = Task.create_task(self._user.username, 'task-1', 'method')
        t2 = Task.create_task(self._user.username, 'task-2', 'method')
        Task.create_task(self._user.username, 'task-3', 'method')
        Task.create_task(self._user.username, 'task-4', 'method')
        Task.create_task(self._user.username, 'task-5', 'method')

        task_list = client.get_tasks(_make_request(), [t1, t2], None)

        self.assertEqual(len(task_list), 2)
        self.assertEqual(set([t['id'] for t in task_list]), set([t1, t2]))

    def test_get_tasks_filter_by_state(self):
        t1 = Task.create_task(self._user.username, 'task-1', 'method', state=TASK_STATES['OPEN'])
        t2 = Task.create_task(self._user.username, 'task-2', 'method', state=TASK_STATES['TIMEOUT'])
        Task.create_task(self._user.username, 'task-3', 'method')
        Task.create_task(self._user.username, 'task-4', 'method')
        Task.create_task(self._user.username, 'task-5', 'method')

        task_list = client.get_tasks(_make_request(), None, [TASK_STATES['OPEN'], TASK_STATES['TIMEOUT']])

        self.assertEqual(len(task_list), 2)
        self.assertEqual(set([t['id'] for t in task_list]), set([t1, t2]))

    def test_cancel_task(self):
        task_id = Task.create_task(self._user.username, 'task', 'method')
        ret = client.cancel_task(_make_request(), task_id)
        self.assertIsNone(ret)

        t = Task.objects.get(id=task_id)
        self.assertEqual(t.state, TASK_STATES['CANCELED'])

    def test_cancel_task_non_existent(self):
        ret = client.cancel_task(_make_request(), 999)
        self.assertEqual(ret, 'Specified task 999 does not exist.')

    def test_resubmit_task(self):
        task_id = Task.create_task(self._user.username, 'task', 'method', state=TASK_STATES['TIMEOUT'])
        new_id = client.resubmit_task(_make_request(self._user), task_id, force=False)
        self.assertTrue(new_id > 0)
        self.assertNotEqual(task_id, new_id)
        task = Task.objects.get(id=new_id)
        self.assertEqual(task.priority, 10)

    def test_resubmit_task_do_not_failed(self):
        task_id = Task.create_task(self._user.username, 'task', 'method', state=TASK_STATES['OPEN'])

        with self.assertRaises(Exception):
            client.resubmit_task(_make_request(), task_id, force=False)

    def test_resubmit_task_force_not_failed(self):
        task_id = Task.create_task(self._user.username, 'task', 'method', state=TASK_STATES['OPEN'])
        new_id = client.resubmit_task(_make_request(self._user), task_id, force=True)
        self.assertTrue(new_id > 0)
        self.assertNotEqual(task_id, new_id)

    def test_resubmit_task_non_existent(self):
        with self.assertRaises(Task.DoesNotExist):
            client.resubmit_task(_make_request(), 999)

    def test_resubmit_task_set_priority(self):
        task_id = Task.create_task(self._user.username, 'task', 'method', state=TASK_STATES['TIMEOUT'])
        new_id = client.resubmit_task(_make_request(self._user), task_id, force=False, priority=19)
        task = Task.objects.get(id=new_id)
        self.assertEqual(task.priority, 19)

    def test_list_workers_enabled(self):
        Worker.objects.create(
            worker_key='enabled-worker',
            name='enabled-worker',
            enabled=True,
        )

        Worker.objects.create(
            worker_key='disabled-worker',
            name='disabled-worker',
            enabled=False,
        )

        worker_list = client.list_workers(_make_request(), True)
        self.assertEqual(len(worker_list), 1)
        self.assertEqual(worker_list[0], 'enabled-worker')

    def test_list_workers_disabled(self):
        Worker.objects.create(
            worker_key='enabled-worker',
            name='enabled-worker',
            enabled=True,
        )

        Worker.objects.create(
            worker_key='disabled-worker',
            name='disabled-worker',
            enabled=False,
        )

        worker_list = client.list_workers(_make_request(), False)
        self.assertEqual(len(worker_list), 1)
        self.assertEqual(worker_list[0], 'disabled-worker')

    def test_create_task(self):
        task_id = client.create_task(_make_request(self._user), {
            'owner_name': self._user.username,
            'label': 'task',
            'method': 'method',
        })

        self.assertTrue(task_id > 0)

        task = Task.objects.get(id=task_id)
        self.assertEqual(task.resubmitted_by.id, self._user.id)
        self.assertIsNone(task.resubmitted_from)

    def test_create_task_based_on_task(self):
        base_task_id = Task.create_task(self._user.username, 'task', 'method')

        clone_task_id = client.create_task(_make_request(self._user), {
            'task_id': base_task_id,
        })

        self.assertTrue(clone_task_id > 0)
        self.assertNotEqual(base_task_id, clone_task_id)

        task = Task.objects.get(id=clone_task_id)
        self.assertEqual(task.resubmitted_by.id, self._user.id)
        self.assertEqual(task.resubmitted_from.id, base_task_id)

    def test_task_url(self):
        url = client.task_url(_make_request(meta={
            'SERVER_PORT': '80',
            'HTTP_HOST': 'example.com',
            'SERVER_NAME': 'srv-name',
        }), 999)

        self.assertEqual(url, 'http://example.com/task/999/')

    def test_task_url_ssl(self):
        url = client.task_url(_make_request(meta={
            'SERVER_PORT': '443',
            'HTTP_HOST': 'example.com',
            'SERVER_NAME': 'srv-name',
        }), 999)

        self.assertEqual(url, 'https://example.com/task/999/')

    def test_task_url_localhost(self):
        url = client.task_url(_make_request(meta={
            'SERVER_PORT': '80',
            'HTTP_HOST': 'localhost',
            'SERVER_NAME': 'srv-name',
        }), 999)

        self.assertEqual(url, 'http://srv-name/task/999/')


class TestXmlRpcClientAuthentication(django.test.TransactionTestCase):

    def test_shutdown_worker_raise_if_no_auth(self):
        with self.assertRaises(PermissionDenied):
            client.shutdown_worker(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 'worker')

    def test_enable_worker_raise_if_no_auth(self):
        with self.assertRaises(PermissionDenied):
            client.enable_worker(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 'worker')

    def test_disable_worker_raise_if_no_auth(self):
        with self.assertRaises(PermissionDenied):
            client.disable_worker(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 'worker')

    def test_cancel_task_raise_if_no_auth(self):
        with self.assertRaises(PermissionDenied):
            client.cancel_task(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 999)

    def test_resubmit_task_raise_if_no_auth(self):
        with self.assertRaises(PermissionDenied):
            client.resubmit_task(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 999)

    def test_create_task_raise_if_no_auth(self):
        with self.assertRaises(PermissionDenied):
            client.create_task(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), {})

    def test_shutdown_worker_raise_if_authenticated_but_not_admin(self):
        with self.assertRaises(PermissionDenied):
            client.shutdown_worker(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 'worker')

    def test_enable_worker_raise_if_authenticated_but_not_admin(self):
        with self.assertRaises(PermissionDenied):
            client.enable_worker(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 'worker')

    def test_disable_worker_raise_if_authenticated_but_not_admin(self):
        with self.assertRaises(PermissionDenied):
            client.disable_worker(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), 'worker')

    def test_create_task_raise_if_authenticated_but_not_admin(self):
        with self.assertRaises(PermissionDenied):
            client.create_task(_make_request(
                is_authenticated=False,
                is_superuser=False,
            ), {})



# -*- coding: utf-8 -*-


from django.conf.urls import url
from kobo.django.views.generic import ExtraListView
from kobo.hub.views import ArchDetailView
from kobo.hub.models import Arch
from kobo.django.compat import gettext_lazy as _


urlpatterns = [
    url(r"^$", ExtraListView.as_view(
        queryset=Arch.objects.order_by("name"),
        template_name="arch/list.html",
        context_object_name="arch_list",
        title=_("Architectures"),
    ), name="arch/list"),
    url(r"^(?P<pk>\d+)/$", ArchDetailView.as_view(), name="arch/detail"),
]



# -*- coding: utf-8 -*-


from kobo.types import Enum


__all__ = (
    "TASK_STATES",
    "FINISHED_STATES",
    "FAILED_STATES",
)


TASK_STATES = Enum(
    "FREE",        # default state for new tasks
    "ASSIGNED",    # assigned to a worker
    "OPEN",        # opened by a worker and being processed
    "CLOSED",      # successfully finished
    "CANCELED",    # canceled by user request
    "FAILED",      # failed
    "INTERRUPTED", # interrupted by an external event (power outage, process killed, etc.)
    "TIMEOUT",     # reached timeout and killed by task manager
    "CREATED",     # task is created, but still not ready to be processed
)


FINISHED_STATES = (
    TASK_STATES["CLOSED"],
    TASK_STATES["CANCELED"],
    TASK_STATES["FAILED"],
    TASK_STATES["INTERRUPTED"],
    TASK_STATES["TIMEOUT"],
)


FAILED_STATES = (
    TASK_STATES["CANCELED"],
    TASK_STATES["FAILED"],
    TASK_STATES["INTERRUPTED"],
    TASK_STATES["TIMEOUT"],
)



from django.contrib.auth.admin import *
import django.contrib.admin as admin
from kobo.django.django_version import django_version_ge

from kobo.django.auth.models import *

# users are not displayed on admin page since migrations were introduced
if django_version_ge("1.9.0"):
    admin.site.register(User, UserAdmin)


#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import os
import subprocess


def test_fields():
    # FIXME: this test should be refactored to work within the current
    # process.  Doing it this way, calling to a subprocess, will effectively
    # disable several features from the test framework.
    subprocess.check_call([sys.executable, __file__])


def main():
    # Runs the test under fields_test directory with custom settings
    PROJECT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
    os.environ["PYTHONPATH"] = PROJECT_DIR
    os.environ['DJANGO_SETTINGS_MODULE'] = 'fields_test.settings'
    sys.path.insert(0, PROJECT_DIR)

    from django.core.management import call_command
    import django

    # Django >= 1.7 must call this method to initialize app registry,
    # while older Django do not have this method
    if 'setup' in dir(django):
        django.setup()

    call_command('test', 'fields_test')
    #call_command('syncdb')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-


import os
import sys
import signal
import optparse
import logging

# IMPORTANT: import taskd first to set os.environ["PROJECT_DEFAULT_CONFIG_FILE"]
from kobo.worker.taskmanager import TaskManager, TaskContainer

import kobo.conf
import kobo.log
import kobo.process
from kobo.exceptions import ShutdownException
from kobo.tback import Traceback, set_except_hook
set_except_hook()


def daemon_shutdown(*args, **kwargs):
    raise ShutdownException()


def main_loop(conf, foreground=False):
    """infinite daemon loop"""

    # define custom signal handlers
    signal.signal(signal.SIGTERM, daemon_shutdown)

    # initialize TaskManager
    try:
        log_file = conf.get("LOG_FILE", None)
        logger = logging.Logger("TaskManager")
        logger.setLevel(logging.DEBUG)
        if log_file:
            log_level = logging.getLevelName(conf.get("LOG_LEVEL", "DEBUG").upper())
            kobo.log.add_rotating_file_logger(logger, log_file, log_level=log_level)

        tm = TaskManager(conf=conf, logger=logger)
    except Exception as ex:
        raise
        sys.stderr.write("Error initializing TaskManager: %s\n" % ex)
        sys.exit(1)

    if foreground and tm._logger is not None:
        kobo.log.add_stderr_logger(tm._logger)

    while 1:
        try:
            tm.log_debug(80 * '-')
            # poll hub for new tasks
            tm.hub._login()
            tm.update_worker_info()
            tm.update_tasks()
            tm.get_next_task()

            # write to stdout / stderr
            sys.stdout.flush()
            sys.stderr.flush()

            # sleep for some time
            tm.sleep()

        except (ShutdownException, KeyboardInterrupt):
            # ignore keyboard interrupts and sigterm
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)

            tm.log_info('Exiting...')
            tm.shutdown()
            tm.log_info('Cleanup done...')
            break

        except:
            # this is a little extreme: log the exception and continue
            traceback = Traceback()
            tm.log_error(traceback.get_traceback())
            tm.sleep()


def main(conf, argv=None):
    parser = optparse.OptionParser()
    parser.add_option(
        "-f", "--foreground",
        default=False,
        action="store_true",
        help="run in foreground (do not spawn a daemon)",
    )
    parser.add_option(
        "-k", "--kill",
        default=False,
        action="store_true",
        help="kill the daemon",
    )
    parser.add_option(
        "-p", "--pid-file",
        help="specify a pid file",
    )
    opts, args = parser.parse_args(argv)

    pid_file = opts.pid_file
    if pid_file is None:
        pid_file = conf.get("PID_FILE")

    if pid_file is None:
        parser.error("No pid file specified.")

    if opts.kill:
        pid = open(pid_file, "r").read()
        os.kill(int(pid), 15)
        sys.exit(0)

    if opts.foreground:
        main_loop(conf, foreground=True)
    else:
        kobo.process.daemonize(main_loop, conf=conf, daemon_pid_file=pid_file, foreground=False)



# -*- coding: utf-8 -*-


from django.conf.urls import url
from kobo.django.views.generic import ExtraListView, ExtraDetailView
from kobo.hub.models import Worker
from kobo.django.compat import gettext_lazy as _


urlpatterns = [
    url(r"^$", ExtraListView.as_view(
        queryset=Worker.objects.order_by("name"),
        template_name="worker/list.html",
        context_object_name="worker_list",
        title = _("Workers"),
    ), name="worker/list"),
    url(r"^(?P<pk>\d+)/$", ExtraDetailView.as_view(
        queryset=Worker.objects.select_related(),
        template_name="worker/detail.html",
        context_object_name="worker",
        title=_("Worker detail"),
    ), name="worker/detail"),
]



# -*- coding: utf-8 -*-


import socket

from django.core.exceptions import PermissionDenied, SuspiciousOperation
from kobo.decorators import decorator_with_args
from kobo.django.helpers import call_if_callable
from kobo.django.xmlrpc.decorators import *


def validate_worker(func):
    def _new_func(request, *args, **kwargs):
        if not call_if_callable(request.user.is_authenticated):
            raise PermissionDenied("Login required.")

        if getattr(request, 'worker', None) is None:
            raise SuspiciousOperation("User doesn't match any worker: %s" % request.user.username)

        request.worker.update_last_seen()

        return func(request, *args, **kwargs)

    _new_func.__name__ = func.__name__
    _new_func.__doc__ = func.__doc__
    _new_func.__dict__.update(func.__dict__)
    return _new_func



import pytest
import gssapi
import mock
import sys

from kobo.xmlrpc import SafeCookieTransport
from kobo.conf import PyConfigParser
from kobo.client import HubProxy


@pytest.fixture(autouse=True)
def requests_session():
    """Mocker for requests.Session; autouse to ensure no accidental real requests.

    Note the tests in this file can't be implemented using requests_mocker because that
    library doesn't track info about authentication.
    """
    with mock.patch("requests.Session") as s:
        # 'with requests.Session()' returns the session instance.
        s.return_value.__enter__.return_value = s.return_value
        yield s


class FakeTransport(SafeCookieTransport):
    """A fake XML-RPC transport where every request succeeds without doing anything.

    Subclasses the real SafeCookieTransport so we get a real CookieJar.
    """
    def __init__(self, *args, **kwargs):
        # note: py2 transport classes do not subclass object
        if sys.version_info[0] < 3:
            SafeCookieTransport.__init__(self, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)

        self.fake_transport_calls = []

    def request(self, host, path, request, verbose=False):
        self.fake_transport_calls.append((path, request))
        return []


def test_login_gssapi(requests_session):
    """Login with gssapi method obtains session cookie via SPNEGO & krb5login."""

    hub_url = "https://example.com/myapp/endpoint"
    login_url = "https://example.com/myapp/auth/krb5login/"

    conf = PyConfigParser()
    conf.load_from_dict(
        {"HUB_URL": hub_url, "AUTH_METHOD": "gssapi",}
    )

    transport = FakeTransport()
    proxy = HubProxy(conf, transport=transport)

    # Proxy might have already done some calls during initialization.
    # We're trying to test login in isolation, so keep track of how many
    # mock calls there have been already.
    mock_get = requests_session.return_value.get
    calls_before = len(mock_get.mock_calls)

    # Force a login
    proxy._login(force=True)

    # Cookies should have been shared between session and transport
    assert requests_session.return_value.cookies is transport.cookiejar

    # Check the requests done
    calls = mock_get.mock_calls[calls_before:]

    assert calls[0][0] == ""
    call_args = calls[0][1]
    call_kwargs = calls[0][2]

    # It should have made a request to log in
    assert call_args == (login_url,)

    # It should have enabled SPNEGO auth.
    # More details about this object are verified in a separate test.
    assert "HTTPSPNEGOAuth" in str(type(call_kwargs["auth"]))

    # It should have verified the result
    assert calls[1][0] == "().raise_for_status"

    # And that's all
    assert len(calls) == 2


def test_login_gssapi_krb_opts(requests_session):
    """Login with gssapi method prepares auth using correct gssapi parameters
    according to config."""

    hub_url = "https://hub.example.com/myapp/endpoint"
    login_url = "https://hub.example.com/myapp/auth/krb5login/"

    conf = PyConfigParser()
    conf.load_from_dict(
        {
            "HUB_URL": hub_url,
            "AUTH_METHOD": "gssapi",
            "CA_CERT": "/some/ca-bundle.pem",
            "KRB_PRINCIPAL": "someclient@EXAMPLE.COM",
            "KRB_SERVICE": "SVC",
            "KRB_REALM": "REALM.EXAMPLE.COM",
            "KRB_KEYTAB": "some-keytab",
            "KRB_CCACHE": "some-cache",
        }
    )

    transport = FakeTransport()
    proxy = HubProxy(conf, transport=transport)

    mock_get = requests_session.return_value.get
    calls_before = len(mock_get.mock_calls)

    with mock.patch("requests_gssapi.HTTPSPNEGOAuth") as mock_auth:
        with mock.patch("gssapi.Credentials") as mock_creds:
            # Force a login
            proxy._login(force=True)

    get_call = mock_get.mock_calls[calls_before]

    # It should have prepared credentials with the details from config
    mock_creds.assert_called_once_with(
        name=gssapi.Name("someclient@EXAMPLE.COM", gssapi.NameType.kerberos_principal),
        store={"client_keytab": "some-keytab", "ccache": "FILE:some-cache"},
        usage="initiate",
    )

    # It should have prepared auth with those credentials and our configured
    # server principal
    mock_auth.assert_called_once_with(
        creds=mock_creds.return_value,
        mutual_authentication=2,
        target_name=gssapi.Name(
            "SVC/hub.example.com@REALM.EXAMPLE.COM", gssapi.NameType.kerberos_principal
        ),
    )

    # It should have used the configured CA bundle when issuing the request
    assert get_call[2]["verify"] == "/some/ca-bundle.pem"


def test_login_gssapi_principal_needs_keytab(requests_session):
    """Login with gssapi method raises if principal is provided without keytab."""
    hub_url = "https://hub.example.com/myapp/endpoint"

    conf = PyConfigParser()
    conf.load_from_dict(
        {
            "HUB_URL": hub_url,
            "AUTH_METHOD": "gssapi",
            "KRB_PRINCIPAL": "someclient@EXAMPLE.COM",
        }
    )

    transport = FakeTransport()
    logger = mock.Mock()
    proxy = HubProxy(conf, transport=transport, logger=logger)

    proxy._login(force=True)

    # This is pretty dumb: login() swallows all exceptions (probably for no good reason).
    # The only hint there was a problem is a DEBUG log message, so we detect the error
    # that way.
    logger.debug.assert_called_with(
        "Failed to create new session: Cannot specify a principal without a keytab"
    )


def test_no_auto_logout(requests_session):
    """auto_logout argument warns of deprecation"""
    conf = PyConfigParser()
    conf.load_from_dict({"HUB_URL": 'https://example.com/hub'})

    transport = FakeTransport()
    with pytest.deprecated_call():
        HubProxy(conf, transport=transport, auto_logout=True)


def test_proxies_to_xmlrpc(requests_session):
    """HubProxy proxies to underlying XML-RPC ServerProxy"""
    conf = PyConfigParser()
    conf.load_from_dict({"HUB_URL": 'https://example.com/hub'})

    transport = FakeTransport()
    proxy = HubProxy(conf, transport=transport)

    proxy.some_obj.some_method()

    # Last call should have invoked the method I requested
    (_, request_xml) = transport.fake_transport_calls[-1]
    assert b'some_obj.some_method' in request_xml



# -*- coding: utf-8 -*-

from __future__ import absolute_import
import os
import datetime
import inspect

from django.core.exceptions import PermissionDenied, SuspiciousOperation
from kobo.decorators import decorator_with_args
from kobo.django.helpers import call_if_callable
from kobo.shortcuts import random_string
from kobo.tback import Traceback

from .models import XmlRpcLog
import six
from six.moves import zip


__all__ = (
    "user_passes_test",
    "login_required",
    "admin_required",
    "validate_user",
    "log_call",
    "log_traceback",
)


@decorator_with_args
def user_passes_test(func, test_func):
    def _new_func(request, *args, **kwargs):
        if not test_func(request.user):
            message = "Permission denied."
            raise PermissionDenied(message)
        return func(request, *args, **kwargs)

    _new_func.__name__ = func.__name__
    _new_func.__doc__ = func.__doc__
    _new_func.__dict__.update(func.__dict__)
    return _new_func


def login_required(func):
    def _new_func(request, *args, **kwargs):
        if not call_if_callable(request.user.is_authenticated):
            raise PermissionDenied("Login required.")
        return func(request, *args, **kwargs)

    _new_func.__name__ = func.__name__
    _new_func.__doc__ = func.__doc__
    _new_func.__dict__.update(func.__dict__)
    return _new_func


def admin_required(func):
    def _new_func(request, *args, **kwargs):
        if not request.user.is_superuser:
            raise PermissionDenied("Admin only.")
        return func(request, *args, **kwargs)

    _new_func.__name__ = func.__name__
    _new_func.__doc__ = func.__doc__
    _new_func.__dict__.update(func.__dict__)
    return _new_func


def validate_user(func):
    def _new_func(request, user_id, *args, **kwargs):
        if request.user.id != user_id:
            raise SuspiciousOperation("UserID doesn't match logged in user.")
        return func(request, user_id, *args, **kwargs)

    _new_func.__name__ = func.__name__
    _new_func.__doc__ = func.__doc__
    _new_func.__dict__.update(func.__dict__)
    return _new_func


def log_call(function):
    def _new_function(request, *args, **kwargs):
        try:
            argspec = inspect.getargspec(function)
            arg_names = argspec[0][1:]
            known_args = list(zip(arg_names, args))
            unknown_args = list(enumerate(args[len(arg_names):]))
            keyword_args = [ (key, value) for key, value in six.iteritems(kwargs) if (key, value) not in known_args ]

            log = XmlRpcLog()
            log.user = request.user
            log.method = function.__name__
            log.args = str(known_args + unknown_args + keyword_args)
            log.save()
        except:
            pass
        return function(request, *args, **kwargs)

    _new_function.__name__ = function.__name__
    _new_function.__doc__ = function.__doc__
    _new_function.__dict__.update(function.__dict__)
    return _new_function


@decorator_with_args
def log_traceback(function, logdir):
    def _new_function(request, *args, **kwargs):
        try:
            result = function(request, *args, **kwargs)
        except Exception as ex:
            # logdir must be absolute path
            if os.path.abspath(logdir) != logdir:
                raise

            if not os.path.isdir(logdir):
                os.makedirs(logdir)

            # create a file name from function name, date, time and some random characters
            file_name = "%s_%s_%s" % (
                function.__name__,
                datetime.datetime.strftime(datetime.datetime.now(), "%Y-%m-%d_%H-%M-%S"),
                random_string(32),
            )

            # create a file with 0600 perms (log can contain sensitive information like passwords)
            file_path = os.path.join(logdir, file_name)
            fd = os.open(file_path, os.O_CREAT | os.O_WRONLY, 0o600)
            os.write(fd, Traceback().get_traceback())
            os.close(fd)
            raise

        return result

    _new_function.__name__ = function.__name__
    _new_function.__doc__ = function.__doc__
    _new_function.__dict__.update(function.__dict__)
    return _new_function



# -*- coding: utf-8 -*-

import sys
import time
import six.moves.urllib.request as urllib2
try:
    import json
except ImportError:
    import simplejson as json

from kobo.client import ClientCommand


MIN_POLL_INTERVAL = 15

class Watch_Log(ClientCommand):
    """displays task logs incrementally"""
    enabled = True


    def options(self):
        self.parser.usage = "%%prog %s task_id" % self.normalized_name

        self.parser.add_option(
            "--type",
            default="stdout.log",
            action="store",
            help="Show log with this name, default is stdout.log"
        )

        self.parser.add_option(
            "--poll",
            default=MIN_POLL_INTERVAL,
            type="int",
            help="Interval how often server should be polled for new info (seconds >= %s)" % MIN_POLL_INTERVAL
        )

        self.parser.add_option(
            "--nowait",
            default=False,
            action="store_true",
            help="Return after fetching current logfile, don't wait until task finishes"
        )


    def run(self, *args, **kwargs):
        if len(args) != 1:
            self.parser.error("Exactly one task id must be specified.")
        try:
            task_id = int(args[0])
        except ValueError:
            self.parser.error("Task ID should be an integer")

        if kwargs['poll'] < MIN_POLL_INTERVAL:
            self.parser.error("Poll interval has to be higher than %s." % MIN_POLL_INTERVAL)

        # HACK: We're presuming, that urls were not touched and that base_url
        # is also url of web ui. As we suppose that also task.urls were not
        # altered it should work.
        url = self.conf['HUB_URL'].replace('/xmlrpc', '') + '/task/%d/log-json/%s?offset=%d'
        offset = 0
        while True:
            data = json.loads(urllib2.urlopen(url % (task_id, kwargs['type'], offset)).read())
            if data['content']:
                sys.stdout.write(data['content'])
                sys.stdout.flush()
            if kwargs['nowait']:
                break
            next_poll = data.get('next_poll')
            if data['task_finished'] == 1 and next_poll is None:
                break
            offset = data['new_offset']

            # If next_poll is 0, that means there's immediately more content, so fetch
            # it now. Otherwise, stick with the user's requested poll interval
            if next_poll != 0:
                time.sleep(kwargs['poll'])



# -*- coding: utf-8 -*-


from django.conf.urls import url
import kobo.hub.views
from kobo.django.django_version import django_version_ge

if django_version_ge('1.11.0'):
    urlpatterns = [
        url(r"^login/$", kobo.hub.views.LoginView.as_view(), name="auth/login"),
        url(r"^krb5login/$", kobo.hub.views.krb5login, name="auth/krb5login"),
        url(r"^oidclogin/$", kobo.hub.views.oidclogin, name="auth/oidclogin"),
        url(r'^logout/$', kobo.hub.views.LogoutView.as_view(), name="auth/logout"),
    ]

else:
    urlpatterns = [
        url(r"^login/$", kobo.hub.views.login, name="auth/login"),
        url(r"^krb5login/$", kobo.hub.views.krb5login, name="auth/krb5login"),
        url(r"^oidclogin/$", kobo.hub.views.oidclogin, name="auth/oidclogin"),
        url(r'^logout/$', kobo.hub.views.logout, name="auth/logout"),
    ]



# -*- coding: utf-8 -*-


# based on: http://code.activestate.com/recipes/146306/


from six.moves import http_client as httplib
import mimetypes
import os

from kobo.shortcuts import random_string


class POSTTransport(object):
    """
    POST transport.

    USAGE:
    >>> import kobo.http
        t = kobo.http.POSTTransport()
        t.add_variable("foo", "bar")
        t.add_file("foofile", "/tmp/some_file")
        t.send_to_host("somehost", "/cgi-bin/upload")
    """

    def __init__(self):
        self._variables = []
        self._files = []
        self._boundary = random_string(32)
        self.last_response = None

    def get_content_type(self, file_name):
        """Guess the mime type of a file.

        @param file_name: file name
        @type file_name: str
        @return: MIME type
        @rtype: str
        """
        return mimetypes.guess_type(file_name)[0] or "application/octet-stream"

    def add_variable(self, key, value):
        """Add a variable to the POST request."""
        self._variables.append((str(key), str(value)))

    def add_file(self, key, file_name):
        """
        Add a file to the POST request.

        @param key: key
        @type key: str
        @param file_name: file name
        @type file_name: str
        """
        if type(file_name) is not str:
            raise TypeError("Invalid type of 'file_name': %s" % type(file_name))

        if not os.path.isfile(file_name):
            raise OSError("Not a file: %s" % file_name)

        self._files.append((str(key), str(file_name)))

    def flush_data(self):
        """Flush variables and files from the request."""
        self._variables = []
        self._files = []

    def send_to_host(self, host, selector, port=None, secure=False, flush=True):
        """
        Send the POST request to a host.

        @param host: host address
        @type host: str
        @param selector: selector/path string
        @type selector: str
        @param port: port number
        @type port: int
        @param secure: use https
        @type secure: bool
        @param flush: flush data after send
        @type flush: bool
        @return: (response status code, response data body)
        @rtype: (int, str)
        """
        content_length = 0

        variables = []
        for key, value in self._variables:
            variables.extend((
                "--%s" % self._boundary,
                'Content-Disposition: form-data; name="%s"' % key,
                "",
                value,
            ))
        variables_data = "\r\n".join(variables)
        variables_data = variables_data.encode()
        content_length += len(variables_data)
        content_length += 2 # '\r\n'

        files = []
        for key, file_name in self._files:
            file_data = "\r\n".join((
                "--%s" % self._boundary,
                'Content-Disposition: form-data; name="%s"; filename="%s"' % (key, os.path.basename(file_name)),
                "Content-Type: %s" % self.get_content_type(file_name),
                "",
                "", # this adds extra newline before file data
            ))
            file_data = file_data.encode()
            files.append((file_name, file_data))
            content_length += len(file_data)
            content_length += os.path.getsize(file_name)
            content_length += 2 # '\r\n'

        footer_data = "\r\n".join(("--%s--" % self._boundary, ""))
        footer_data = footer_data.encode()
        content_length += len(footer_data)
        content_type = "multipart/form-data; boundary=" + self._boundary

        if secure:
            request = httplib.HTTPSConnection(host, port)
        else:
            request = httplib.HTTPConnection(host, port)

        request.putrequest("POST", selector)
        request.putheader("content-type", content_type)
        request.putheader("content-length", str(content_length))
        request.endheaders()
        request.send(variables_data)
        request.send(b"\r\n")
        for file_name, file_data in files:
            request.send(file_data)
            file_obj = open(file_name, "rb")
            while 1:
                chunk = file_obj.read(1024**2)
                if not chunk:
                    break
                request.send(chunk)
            request.send(b"\r\n")

        request.send(footer_data)
        response = request.getresponse()

        if flush:
            self.flush_data()

        return response.status, response.read()



# -*- coding: utf-8 -*-

import unittest

from mock import Mock, PropertyMock

from kobo.exceptions import ShutdownException
from kobo.worker.tasks.task_shutdown_worker import ShutdownWorker

class TestShutdownWorker(unittest.TestCase):

    def test_run(self):
        t = ShutdownWorker(Mock(spec=['worker']), {}, 100, {})

        t.task_manager = PropertyMock(locked=False)
        self.assertFalse(t.task_manager.locked)

        t.run()
        self.assertTrue(t.task_manager.locked)

    def test_run_kill(self):
        t = ShutdownWorker(Mock(spec=['worker']), {}, 100, {'kill': True})

        t.task_manager = PropertyMock(locked=False)
        self.assertFalse(t.task_manager.locked)

        with self.assertRaises(ShutdownException):
            t.run()

        self.assertFalse(t.task_manager.locked)



# -*- coding: utf-8 -*-

import time
import logging

import unittest

from six import BytesIO, StringIO

from mock import Mock

from kobo.worker.logger import LoggingThread, LoggingIO
from kobo.log import LoggingBase
from .utils import ArgumentIsInstanceOf


class TestLoggingThread(unittest.TestCase):

    def test_upload_task_log_on_stop(self):
        mock_hub = Mock()
        thread = LoggingThread(mock_hub, 9999)
        thread.daemon = True

        thread.start()
        self.assertTrue(thread.is_alive())
        self.assertTrue(thread._running)

        thread.write('This is a log message!')
        mock_hub.upload_task_log.assert_not_called()

        thread.stop()
        self.assertFalse(thread.is_alive())
        self.assertFalse(thread._running)
        mock_hub.upload_task_log.assert_called_once_with(ArgumentIsInstanceOf(BytesIO), 9999, 'stdout.log', append=True)

    def test_upload_task_log_after_some_time(self):
        mock_hub = Mock()
        thread = LoggingThread(mock_hub, 9999)
        thread.daemon = True

        thread.start()

        self.assertTrue(thread.is_alive())
        self.assertTrue(thread._running)

        for i in range(5):
            thread.write('%d - This is a very long message to be written in the log\n' % (i + 1))
            mock_hub.upload_task_log.assert_not_called()

        # let the thread running for a while
        time.sleep(.1)
        mock_hub.upload_task_log.assert_called_once_with(ArgumentIsInstanceOf(BytesIO), 9999, 'stdout.log', append=True)

        thread.stop()
        self.assertFalse(thread.is_alive())
        self.assertFalse(thread._running)

    def test_logs_on_fatal_error(self):
        # Set up a logger whose output we'll be able to inspect.
        logs = StringIO()
        logger = logging.getLogger('TestLoggingThread')
        logger.addHandler(logging.StreamHandler(logs))
        kobo_logger = LoggingBase(logger)

        # Set up hub to raise some exception other than an XML-RPC fault.
        mock_hub = Mock()
        mock_hub.upload_task_log.side_effect = RuntimeError("Simulated error")

        thread = LoggingThread(mock_hub, 9999, logger=kobo_logger)
        thread.daemon = True
        thread.start()

        thread.write('This is a log message!')

        # Since we set up a fatal error, we expect the thread to die soon
        # despite not calling stop().
        thread.join(10.0)
        self.assertFalse(thread.is_alive())

        # Before dying, it should have written something useful to the logs.
        captured = logs.getvalue()
        assert 'Fatal error in LoggingThread' in captured
        assert 'RuntimeError: Simulated error' in captured

    def test_mixed_writes(self):
        uploaded = []

        # Used as a side-effect to keep whatever's uploaded
        def mock_upload(io, *args, **kwargs):
            uploaded.append(io.read())

        mock_hub = Mock()
        mock_hub.upload_task_log.side_effect = mock_upload

        thread = LoggingThread(mock_hub, 9999)
        thread.daemon = True

        thread.start()

        # Mixing text, ascii-safe bytes, ascii-unsafe bytes here
        # as is possible with sys.stdout at least on py2.
        thread.write('Some text!')
        thread.write(b'Some bytes!')
        thread.write(u'Some 文!')
        thread.write(b'Some \xe2 wacky bytes!')

        # It should be able to stop normally
        thread.stop()
        self.assertFalse(thread.is_alive())

        # It should have uploaded exactly the expected bytes.
        assert b''.join(uploaded) == (
            b'Some text!Some bytes!'
            b'Some \xe6\x96\x87!Some \xe2 wacky bytes!'
        )


class TestLoggingIO(unittest.TestCase):

    def test_write(self):
        mock_io = Mock()
        mock_logging_thread = Mock()

        stream = LoggingIO(mock_io, mock_logging_thread)
        stream.write('This is a log message!')

        mock_io.write.assert_called_once_with('This is a log message!')
        mock_logging_thread.write.assert_called_once_with('This is a log message!')

    def test_write_multiple_calls(self):
        mock_io = Mock()
        mock_logging_thread = Mock()

        stream = LoggingIO(mock_io, mock_logging_thread)

        for i in range(10):
            stream.write('This is the %d log message!' % (i + 1))

        mock_io.write.assert_any_call('This is the 1 log message!')
        mock_io.write.assert_any_call('This is the 10 log message!')

        mock_logging_thread.write.assert_any_call('This is the 1 log message!')
        mock_logging_thread.write.assert_any_call('This is the 10 log message!')

        self.assertEqual(mock_io.write.call_count, 10)
        self.assertEqual(mock_logging_thread.write.call_count, 10)

    def test_getattr(self):
        mock_io = Mock()
        mock_logging_thread = Mock()

        stream = LoggingIO(mock_io, mock_logging_thread)
        random_variable = stream.random_variable

        self.assertFalse(random_variable is None)
        self.assertIsInstance(random_variable, Mock)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import unittest
import os
import tempfile
import shutil

from kobo.client import BaseClientCommandContainer, ClientCommandContainer
from kobo.cli import CommandOptionParser
from kobo.conf import PyConfigParser

TEST_CONFIG = '''
HUB_URL = "https://localhost/hub/xmlrpc"

AUTH_METHOD = "krbv"
'''


class TestBaseClientCommandContainer(unittest.TestCase):
    def setUp(self):
        self.command_container = BaseClientCommandContainer()

    def test_profile_option_unset(self):
        parser = CommandOptionParser(command_container=self.command_container)
        option = parser.get_option("--profile")

        self.assertEqual(parser.default_profile, "")
        self.assertEqual(option, None)

    def test_profile_option_set(self):
        parser = CommandOptionParser(command_container=self.command_container, default_profile="default-profile")
        option = parser.get_option("--profile")

        self.assertEqual(parser.default_profile, "default-profile")
        self.assertEqual(option.get_opt_string(), "--profile")
        self.assertEqual(option.help, "specify profile (default: default-profile)")

    def test_configuration_directory_option_unset(self):
        parser = CommandOptionParser(command_container=self.command_container, default_profile="default-profile")
        # CommandOptionParser() doesn't store the configuration_file path in an instance variable, instead it's
        # build in _load_profile() with the line below:
        configuration_file = os.path.join(parser.configuration_directory, '{0}.conf'.format(parser.default_profile))

        self.assertEqual(parser.configuration_directory, "/etc")
        self.assertEqual(configuration_file, "/etc/default-profile.conf")

    def test_configuration_directory_option_set(self):
        parser = CommandOptionParser(command_container=self.command_container, default_profile="default-profile",
                                     configuration_directory="/etc/client")

        configuration_file = os.path.join(parser.configuration_directory, '{0}.conf'.format(parser.default_profile))

        self.assertEqual(parser.configuration_directory, "/etc/client")
        self.assertEqual(configuration_file, "/etc/client/default-profile.conf")


class TestClientCommandContainer(unittest.TestCase):
    def setUp(self):
        self.dir = tempfile.mkdtemp()
        self.conf = PyConfigParser()

        self.file = os.path.join(self.dir, 'test.conf')

        with open(self.file, 'w') as f:
            f.write(TEST_CONFIG)

        self.conf.load_from_file(self.file)

    def tearDown(self):
        shutil.rmtree(self.dir)

    def test_config_from_file(self):
        container = ClientCommandContainer(self.conf)

        values = {
            'HUB_URL': 'https://localhost/hub/xmlrpc',
            'AUTH_METHOD': 'krbv'
        }
        self.assertEqual(container.conf, values)

    def test_config_from_kwargs(self):
        container = ClientCommandContainer(self.conf, USERNAME='testuser')

        values = {
            'HUB_URL': 'https://localhost/hub/xmlrpc',
            'AUTH_METHOD': 'krbv',
            'USERNAME': 'testuser'
        }
        self.assertEqual(container.conf, values)



# -*- coding: utf-8 -*-

import mimetypes
import os
import six
import locale
from kobo.django.django_version import django_version_ge

try:
    import json
except ImportError:
    import simplejson as json

import django.contrib.auth.views
from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME, get_user_model
from django.core.exceptions import ImproperlyConfigured
from kobo.django.django_version import django_version_ge
if django_version_ge('1.10.0'):
    from django.urls import reverse
else:
    from django.core.urlresolvers import reverse
from django.http import HttpResponse, StreamingHttpResponse, HttpResponseForbidden
from django.shortcuts import render, get_object_or_404
from django.template import RequestContext
from django.views.generic import RedirectView

from kobo.hub.models import Arch, Channel, Task
from kobo.hub.forms import TaskSearchForm
from kobo.django.views.generic import ExtraDetailView, SearchView
from kobo.django.compat import gettext_lazy as _


# max log size returned in HTML-embedded view
HTML_LOG_MAX_SIZE = getattr(settings, "HTML_LOG_MAX_SIZE", (1024 ** 2) * 2)

# max log size returned in a JSON request
JSON_LOG_MAX_SIZE = getattr(settings, "JSON_LOG_MAX_SIZE", (1024 ** 2) * 8)

# default LogWatcher JS poll interval
LOG_WATCHER_INTERVAL = getattr(settings, "LOG_WATCHER_INTERVAL", 5000)


class UserDetailView(ExtraDetailView):
    model = get_user_model()
    title = _("User detail")
    template_name = "user/detail.html"
    context_object_name = "usr"

    def get_context_data(self, **kwargs):
        context = super(UserDetailView, self).get_context_data(**kwargs)
        context['tasks'] = kwargs['object'].task_set.count()
        return context

class DetailViewWithWorkers(ExtraDetailView):
    model = Channel
    template_name = "channel/detail.html"
    context_object_name = "channel"
    title = _("Architecture detail")

    def get_context_data(self, **kwargs):
        context = super(DetailViewWithWorkers, self).get_context_data(**kwargs)
        context["worker_list"] = kwargs["object"].worker_set.order_by("name")
        return context

class ArchDetailView(ExtraDetailView):
    model = Arch
    template_name = "arch/detail.html"
    context_object_name = "arch"
    title = _("Architecture detail")

    def get_context_data(self, **kwargs):
        context = super(ArchDetailView, self).get_context_data(**kwargs)
        context["worker_list"] = kwargs["object"].worker_set.order_by("name")
        return context

class TaskListView(SearchView):
    # TODO: missing kwargs custom queries for backward compatibility
    title = _("All tasks")
    model = Task
    form_class = TaskSearchForm
    template_name = "task/list.html"
    context_object_name = "task_list"
    state = None
    order_by = ['-id']

    def get_form_kwargs(self):
        kwargs = super(TaskListView, self).get_form_kwargs()
        kwargs['state'] = self.state
        kwargs['order_by'] = self.order_by
        return kwargs


class TaskDetail(ExtraDetailView):
    queryset = Task.objects.select_related()
    context_object_name = "task"
    template_name = "task/detail.html"
    title = _("Task detail")

    def get_context_data(self, **kwargs):
        context = super(TaskDetail, self).get_context_data(**kwargs)
        logs = []
        for i in kwargs['object'].logs.list:
            if self.request.user.has_perm('hub.can_see_traceback'):
                logs.append(i)
                continue
            if not os.path.basename(i).startswith("traceback"):
                logs.append(i)
        logs.sort()
        context["logs"] = logs
        context['task_list'] = kwargs['object'].subtasks()
        return context


def _stream_file(file_path, offset=0):
    """Generator that returns 1M file chunks."""
    try:
        f = open(file_path, "rb")
    except IOError:
        return

    f.seek(offset)
    while 1:
        data = f.read(1024 ** 2)
        if not data:
            break
        yield data
    f.close()


def _trim_log(text):
    # break at first line if possible
    nl = text.find('\n')
    if nl > 0:
        subtext = text[nl:]
    else:
        subtext = '\n' + text
    return '<...trimmed, download required for full log>' + subtext


def _streamed_log_response(file_path, offset, as_attachment):
    mimetype = mimetypes.guess_type(file_path)[0] or 'application/octet-stream'

    try:
        content_len = os.path.getsize(file_path) - offset
    except OSError:
        content_len = 0

    # use _stream_file() instead of passing file object in order to improve performance
    response = StreamingHttpResponse(_stream_file(file_path, offset), content_type=mimetype)
    response["Content-Length"] = content_len

    if as_attachment:
        # set filename to be real filesystem name
        response['Content-Disposition'] = 'attachment; filename=%s' % os.path.basename(file_path)

    return response


def _rendered_log_response(request, task, log_name):
    exts = getattr(settings, "VALID_TASK_LOG_EXTENSIONS", [".log"])
    found = False
    for ext in exts:
        if log_name.endswith(ext):
            found = True
    if not found:
        return HttpResponseForbidden("Can display only specific file types: %s" % ", ".join(exts))

    (content, offset) = task.logs.tail(log_name, HTML_LOG_MAX_SIZE)

    # Add "trimmed" message if tail did not return entire log file.
    if len(content) < offset:
        content = b'<...trimmed, download required for full log>\n' + content

    content = content.decode("utf-8", "replace")

    context = {
        "title": "Task log",
        "offset": offset,
        "task_finished": task.is_finished() and 1 or 0,
        "next_poll": None if task.is_finished() else LOG_WATCHER_INTERVAL,
        "content": content,
        "log_name": log_name,
        "task": task,
        "json_url": reverse("task/log-json", args=[task.id, log_name]),
    }

    return render(request, "task/log.html", context)


def task_log(request, id, log_name):
    """
    IMPORTANT: reverse to 'task/log-json' *must* exist
    """
    if os.path.basename(log_name).startswith("traceback") and not request.user.has_perm('hub.can_see_traceback'):
        return HttpResponseForbidden("You don't have permission to see the traceback.")

    task = get_object_or_404(Task, id=id)

    file_path = task.logs._get_absolute_log_path(log_name)
    if not os.path.isfile(file_path) and not file_path.endswith(".gz"):
        file_path = task.logs._get_absolute_log_path(log_name + ".gz")

    offset = int(request.GET.get("offset", 0))

    request_format = request.GET.get("format")

    if request_format == "raw" or log_name.endswith(".html") or log_name.endswith(".htm"):
        return _streamed_log_response(file_path, offset, as_attachment=(request_format == 'raw'))

    return _rendered_log_response(request, task, log_name)


def task_log_json(request, id, log_name):
    if os.path.basename(log_name).startswith("traceback") and not request.user.is_superuser:
        return HttpResponseForbidden(content_type="application/json")

    task = get_object_or_404(Task, id=id)
    offset = int(request.GET.get("offset", 0))
    content = task.logs.get_chunk(log_name, offset, JSON_LOG_MAX_SIZE + 5)

    if len(content) > JSON_LOG_MAX_SIZE:
        # We immediately have more log content to read
        next_poll = 0
        content = content[:JSON_LOG_MAX_SIZE]
    elif task.is_finished():
        # There is certainly nothing more to read
        next_poll = None
    else:
        # Task is not finished, so there might be more to read,
        # check back soon
        next_poll = LOG_WATCHER_INTERVAL

    if six.PY3:
        content = str(content, encoding=locale.getpreferredencoding())

    result = {
        "new_offset": offset + len(content),
        "task_finished": task.is_finished() and 1 or 0,
        "next_poll": next_poll,
        "content": content,
    }

    return HttpResponse(json.dumps(result), content_type="application/json")

if django_version_ge('1.11.0'):
    class LoginView(django.contrib.auth.views.LoginView):
        template_name = 'auth/login.html'

    class LogoutView(django.contrib.auth.views.LogoutView):
        pass

else:
    def login(request, redirect_field_name=REDIRECT_FIELD_NAME):
        return django.contrib.auth.views.login(request, template_name="auth/login.html", redirect_field_name=redirect_field_name)

    def logout(request, redirect_field_name=REDIRECT_FIELD_NAME):
        return django.contrib.auth.views.logout(request, redirect_field_name=redirect_field_name)


def krb5login(request, redirect_field_name=REDIRECT_FIELD_NAME):
    #middleware = 'django.contrib.auth.middleware.RemoteUserMiddleware'
    middleware = 'kobo.django.auth.middleware.LimitedRemoteUserMiddleware'
    if django_version_ge('1.10.0'):
        middleware_setting = settings.MIDDLEWARE
    else:
        middleware_setting = settings.MIDDLEWARE_CLASSES
    if middleware not in middleware_setting:
        raise ImproperlyConfigured("krb5login view requires '%s' middleware installed" % middleware)
    redirect_to = request.POST.get(redirect_field_name, "")
    if not redirect_to:
        redirect_to = request.GET.get(redirect_field_name, "")
    if not redirect_to:
        redirect_to = reverse("home/index")
    return RedirectView.as_view(url=redirect_to, permanent=True)(request)

def oidclogin(request):
    middleware = 'kobo.django.auth.middleware.LimitedRemoteUserMiddleware'
    if django_version_ge('1.10.0'):
        middleware_setting = settings.MIDDLEWARE
    else:
        middleware_setting = settings.MIDDLEWARE_CLASSES
    if middleware not in middleware_setting:
        raise ImproperlyConfigured("oidclogin view requires '%s' middleware installed" % middleware)
    return RedirectView.as_view(url=reverse("home/index"), permanent=False)(request)



# -*- coding: utf-8 -*-

import django

from django.core.exceptions import PermissionDenied
from mock import Mock, patch

from kobo.hub.models import Worker
from kobo.hub.xmlrpc import auth
from kobo.django.django_version import django_version_ge

from .utils import DjangoRunner

runner = DjangoRunner()
setup_module = runner.start
teardown_module = runner.stop


class TestLoginWorker(django.test.TransactionTestCase):

    def test_login_worker_key_valid_worker_and_user(self):
        def login(request, user):
            request.session.session_key = '1234567890'
            return user

        Worker.objects.create(worker_key='key', name='name')

        req = Mock(spec=['session'], session=Mock())
        user = Mock()
        krb_mock = Mock(spec=['authenticate'], authenticate=Mock(return_value=user))

        with patch('kobo.hub.xmlrpc.auth.Krb5RemoteUserBackend', return_value=krb_mock):
            with patch.object(auth.django.contrib.auth, 'login', side_effect=login) as login_mock:
                session_key = auth.login_worker_key(req, 'key')

                login_mock.assert_called_once_with(req, user)
        if django_version_ge('1.11.0'):
            krb_mock.authenticate.assert_called_once_with(None, 'worker/name')
        else:
            krb_mock.authenticate.assert_called_once_with('worker/name')
        self.assertEqual(session_key, '1234567890')

    def test_login_worker_key_valid_worker_invalid_user(self):
        Worker.objects.create(worker_key='key', name='name')
        req = Mock(spec=['session'], session=Mock())
        krb_mock = Mock(spec=['authenticate'], authenticate=Mock(return_value=None))

        with patch('kobo.hub.xmlrpc.auth.Krb5RemoteUserBackend', return_value=krb_mock):
            with self.assertRaises(PermissionDenied):
                auth.login_worker_key(req, 'key')

        if django_version_ge('1.11.0'):
            krb_mock.authenticate.assert_called_once_with(None, 'worker/name')
        else:
            krb_mock.authenticate.assert_called_once_with('worker/name')

    def test_login_worker_key_invalid_worker(self):
        req = Mock()

        with self.assertRaises(PermissionDenied):
            auth.login_worker_key(req, 'key')



# -*- coding: utf-8 -*-


"""
IMPORTANT: Avoid cycles in container inheritance.

If you need multiple inheritance, never do this:
  - class A(PluginContainer)
  - class B(A)
  - class C(A)
  - class D(B, C) # wrong! both C and B inherit from A; this ends with TypeError: Cannot create a consistent method resolution order (MRO)...

Always make sure all trees inherit from PluginContainer:
  - class A(PluginContainer)
  - class B(A)
  - class C(PluginContainer)
  - class D(C)
  - class E(C, D) # correct, the only common predecesor is PluginContainer
"""


from __future__ import print_function
import os
import six


__all__ = (
    "Plugin",
    "PluginContainer",
)


class Plugin(object):
    """A plugin base class."""

    author = None
    version = None
    enabled = False

    def __getattr__(self, name):
        """
        Get missing attribute from a container.
        This is quite hackish but it allows to define settings and methods per container.
        """
        return getattr(self.container, name)


class PluginContainer(object):
    """A plugin container.

    Usage: Inherit PluginContainer and register plugins to the new class.
    """

    def __getitem__(self, name):
        return self._get_plugin(name)

    def __iter__(self):
        return iter(self.plugins)

    @classmethod
    def normalize_name(cls, name):
        return name

    @classmethod
    def _get_plugins(cls):
        """Return dictionary of registered plugins."""

        result = {}
        parent_plugins = list(cls._get_parent_plugins(cls.normalize_name).items())
        class_plugins = list(getattr(cls, "_class_plugins", {}).items())
        for name, plugin_class in parent_plugins + class_plugins:
            result[name] = type(plugin_class.__name__, (plugin_class, ), {"__doc__": plugin_class.__doc__})
        return result

    @classmethod
    def _get_parent_plugins(cls, normalize_function):
        result = {}
        for parent in cls.__bases__:
            if parent is PluginContainer:
                # don't use PluginContainer itself - plugins have to be registered to subclasses
                continue

            if not issubclass(parent, PluginContainer):
                # skip parents which are not PluginContainer subclasses
                continue

            # read inherited plugins first (conflicts are resolved recursively)
            plugins = parent._get_parent_plugins(normalize_function)

            # read class plugins, override inherited on name conflicts
            if hasattr(parent, "_class_plugins"):
                for plugin_class in parent._class_plugins.values():
                    normalized_name = normalize_function(plugin_class.__name__)
                    plugins[normalized_name] = plugin_class

            for name, value in six.iteritems(plugins):
                if result.get(name, value) != value:
                   raise RuntimeError("Cannot register plugin '%s'. Another plugin with the same normalized name (%s) is already in the container." % (str(value), normalized_name))

            result.update(plugins)

        return result

    @property
    def plugins(self):
        if not hasattr(self, "_plugins"):
            self._plugins = self.__class__._get_plugins()
        return self._plugins

    def _get_plugin(self, name):
        """Return a plugin or raise KeyError."""
        normalized_name = self.normalize_name(name)

        if normalized_name not in self.plugins:
            raise KeyError("Plugin not found: %s" % normalized_name)

        plugin = self.plugins[normalized_name]
        plugin.container = self
        plugin.normalized_name = normalized_name
        return plugin

    @classmethod
    def register_plugin(cls, plugin):
        """Register a new plugin. Return normalized plugin name."""

        if cls is PluginContainer:
            raise TypeError("Can't register plugin to the PluginContainer base class.")

        if "_class_plugins" not in cls.__dict__:
            cls._class_plugins = {}

        if not getattr(plugin, "enabled", False):
            return

        normalized_name = cls.normalize_name(plugin.__name__)
        cls._class_plugins[normalized_name] = plugin
        return normalized_name

    @classmethod
    def register_module(cls, module, prefix=None, skip_broken=False):
        """Register all plugins in a module's sub-modules.

        @param module: a python module that contains plugin sub-modules
        @type  module: module
        @param prefix: if specified, only modules with this prefix will be processed
        @type  prefix: str
        @param skip_broken: skip broken sub-modules and print a warning
        @type  skip_broken: bool
        """
        path = os.path.dirname(module.__file__)
        module_list = []

        for fn in os.listdir(path):
            if not fn.endswith(".py"):
                continue
            if fn.startswith("_"):
                continue
            if prefix and not fn.startswith(prefix):
                continue
            if not os.path.isfile(os.path.join(path, fn)):
                continue
            module_list.append(fn[:-3])

        if skip_broken:
            for mod in module_list[:]:
                try:
                    __import__(module.__name__, {}, {}, [mod])
                except:
                    import sys
                    print("WARNING: Skipping broken plugin module: %s.%s" % (module.__name__, mod), file=sys.stderr)
                    module_list.remove(mod)
        else:
            __import__(module.__name__, {}, {}, module_list)

        for mn in module_list:
            mod = getattr(module, mn)
            for pn in dir(mod):
                plugin = getattr(mod, pn)
                if type(plugin) is type and issubclass(plugin, Plugin) and plugin is not Plugin:
                    cls.register_plugin(plugin)



# -*- coding: utf-8 -*-


import os

from kobo.exceptions import ImproperlyConfigured
from django.conf import settings
from kobo.django.django_version import django_version_ge


for var in ["XMLRPC_METHODS", "TASK_DIR", "UPLOAD_DIR"]:
    if not hasattr(settings, var):
        raise ImproperlyConfigured("'%s' is missing in project settings. It must be set to run kobo.hub app." % var)


if not hasattr(settings, "WORKER_DIR"):
    # This setting introduced in 2021 can be defaulted to ensure backwards compatibility
    # with existing config files.
    worker_dir = os.path.join(os.path.dirname(settings.TASK_DIR), 'worker')
    setattr(settings, "WORKER_DIR", worker_dir)


for var in ["TASK_DIR", "UPLOAD_DIR", "WORKER_DIR"]:
    dir_path = getattr(settings, var)
    if not os.path.isdir(dir_path):
        try:
            os.makedirs(dir_path)
        except:
            raise ImproperlyConfigured("'%s' doesn't exist and can't be automatically created." % dir_path)
    elif not os.access(dir_path, os.R_OK | os.W_OK | os.X_OK):
        raise ImproperlyConfigured("Invalid permissions on '%s'." % dir_path)


if getattr(settings, "MIDDLEWARE", None) is not None:
    # Settings defines Django>=1.10 style middleware, check that
    middleware_var = "MIDDLEWARE"
else:
    # Legacy
    middleware_var = "MIDDLEWARE_CLASSES"

for var, value in [(middleware_var, "kobo.hub.middleware.WorkerMiddleware")]:
    if value not in (getattr(settings, var, None) or []):
        raise ImproperlyConfigured("'%s' in '%s' is missing in project settings. It must be set to run kobo.hub app." % (value, var))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import pytest

from shutil import rmtree

import django.conf
import django.test
from django.test.utils import get_runner
from django.shortcuts import get_object_or_404

from kobo.hub.models import Task, Arch, Channel
from django.contrib.auth.models import User

from .utils import DjangoRunner

runner = DjangoRunner()
setup_module = runner.start
teardown_module = runner.stop

TASK_ID = 123

class TestTaskLogs(django.test.TransactionTestCase):
    def cleanDataDirs(self):
        """Delete the log directory for the test task."""
        log_dir = os.path.join(django.conf.settings.TASK_DIR, '0', '0', str(TASK_ID))
        if os.path.exists(log_dir):
            rmtree(log_dir)

    def tearDown(self):
        self.cleanDataDirs()

        super(TestTaskLogs, self).tearDown()

    def setUp(self):
        self._fixture_teardown()
        super(TestTaskLogs, self).setUp()

        self.cleanDataDirs()

        # These objects are required to exist but their attributes are irrelevant
        user = User.objects.create()
        arch = Arch.objects.create(name='testarch')
        channel = Channel.objects.create(name='testchannel')

        test_task = Task.objects.create(
            pk=TASK_ID, arch=arch, channel=channel, owner=user)

        self.log_content = log_content = b'Line 1 \xe2\x98\xba\nLine 2 \xe2\x98\xba\nLine 3'

        test_task.logs['test_compressed.log'] = log_content
        test_task.logs.save()
        test_task.logs.gzip_logs()

        test_task.logs['test.log'] = log_content
        test_task.logs.save()

    def task_logs(self):
        return get_object_or_404(Task, id=TASK_ID).logs

    def test_read_log_typical(self):
        task_logs = self.task_logs()
        log_content = task_logs['test.log']
        self.assertEqual(log_content, self.log_content)

        # It should be in the cache
        self.assertEqual(log_content, task_logs.cache['test.log'])

    def test_read_log_compressed(self):
        task_logs = self.task_logs()

        log_content = task_logs['test_compressed.log']
        self.assertEqual(log_content, self.log_content)

        # It should be in the cache
        self.assertEqual(log_content, task_logs.cache['test_compressed.log'])

    def test_get_chunk_missing(self):
        do_call = lambda: self.task_logs().get_chunk('notexist.log')
        self.assertRaises(Exception, do_call)

    def test_get_chunk_all(self):
        self.assertEqual(self.task_logs().get_chunk('test.log'), self.log_content)

    def test_get_chunk_offset(self):
        chunk = self.task_logs().get_chunk('test.log', offset=len(b'Line 1 \xe2\x98\xba\n'))
        self.assertEqual(chunk, b'Line 2 \xe2\x98\xba\nLine 3')

    def test_get_chunk_offset_len(self):
        self.do_chunk_offset_len_test('test.log')
        self.do_chunk_offset_len_test('test_compressed.log')

    def do_chunk_offset_len_test(self, log_name):
        chunk = self.task_logs().get_chunk(log_name, offset=1, length=2)
        self.assertEqual(chunk, b'in')

    def test_chunk_offset_len_from_cache(self):
        task_logs = self.task_logs()

        # First, read the entire log.
        self.assertEqual(self.log_content, task_logs['test.log'])

        # Now, if I request a chunk, it comes directly from the cache
        # rather than reading from disk again, proven by deleting
        # the log dir before requesting chunk
        self.cleanDataDirs()

        chunk = task_logs.get_chunk('test.log', offset=1, length=2)
        self.assertEqual(chunk, b'in')

    def test_get_chunk_offset_within_char(self):
        self.do_chunk_offset_within_char_test('test.log')
        self.do_chunk_offset_within_char_test('test_compressed.log')

    def do_chunk_offset_within_char_test(self, log_name):
        task_logs = self.task_logs()
        offset = self.log_content.index(b'\n') + 1
        expected_chunk = b'Line 2 \xe2\x98\xba'
        byteslen = len(expected_chunk)

        # ☺ character requires three bytes to encode; see what happens if we try to read
        # within that char

        full_chunk = task_logs.get_chunk(log_name, offset=offset, length=byteslen)
        chunk_min1 = task_logs.get_chunk(log_name, offset=offset, length=byteslen-1)
        chunk_min2 = task_logs.get_chunk(log_name, offset=offset, length=byteslen-2)
        chunk_min3 = task_logs.get_chunk(log_name, offset=offset, length=byteslen-3)

        # reading entirety of character should be fine
        self.assertEqual(full_chunk, expected_chunk)

        # reading prior to character should be fine
        self.assertEqual(chunk_min3, b'Line 2 ')

        # these two requests to read within the character should have truncated and
        # given the same result as reading prior to the character
        self.assertEqual(chunk_min1, chunk_min3)
        self.assertEqual(chunk_min2, chunk_min3)



#!/usr/bin/python
# -*- coding: utf-8 -*-


import unittest

import tempfile
import os.path
import shutil

from kobo.hardlink import Hardlink, UndoHardlink


class LoggerMock(object):
    def __init__(self):
        self.loglvl = None
        self.msg = None

    def log(self, loglvl, msg):
        self.loglvl = loglvl
        self.msg = msg


class TestHardlinkClass(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_link(self):
        path_src = os.path.join(self.tmp_dir, "a")
        open(path_src, 'w').write("asdf")
        path_dst = os.path.join(self.tmp_dir, "b")

        hl = Hardlink()
        hl.link(path_src, path_dst)

        self.assertEqual(os.stat(path_src),  os.stat(path_dst))

    def test_log(self):
        logger = LoggerMock()
        hl = Hardlink(logger=logger)
        log_level = 1
        message = "foobar"
        hl.log(log_level, message)

        self.assertEqual(logger.loglvl, log_level)
        self.assertEqual(logger.msg, message)


class TestUndoHardlinkClass(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_undo_hardlink(self):
        path_src = os.path.join(self.tmp_dir, "a")
        path_dst = os.path.join(self.tmp_dir, "b")
        open(path_src, 'w').write("asdf")
        old_stat = os.stat(path_src)

        # Only one hardlink exists
        uhl1 = UndoHardlink()
        uhl1.undo_hardlink(path_src)
        new_stat = os.stat(path_src)
        self.assertEqual(old_stat.st_nlink, new_stat.st_nlink)

        # Two hardlinks exist
        os.link(path_src, path_dst)
        uhl2 = UndoHardlink()
        uhl2.undo_hardlink(path_dst)
        new_stat = os.stat(path_dst)
        self.assertNotEqual(old_stat.st_ino, new_stat.st_ino)
        self.assertEqual(new_stat.st_nlink, 1) # Expected num of hardlinks is 1



try:
    # Ancient case: ugettext is the unicode-aware variant
    from django.utils.translation import ugettext_lazy as gettext_lazy
except ImportError:
    # Modern (py3-only) case
    from django.utils.translation import gettext_lazy


try:
    # Ancient case: force_text is the unicode-aware variant
    from django.utils.encoding import force_text as force_str
except:
    # Modern (py3-only) case
    from django.utils.encoding import force_str


__all__ = ["gettext_lazy", "force_str"]



# -*- coding: utf-8 -*-


import errno
import os
import logging
import logging.handlers
import warnings


__all__ = (
    "BRIEF_LOG_FORMAT",
    "VERBOSE_LOG_FORMAT",
    "add_stream_handler",
    "add_stderr_logger",
    "add_file_logger",
    "add_rotating_file_logger",
    "LoggingBase",
)


BRIEF_LOG_FORMAT = "%(asctime)s [%(levelname)-8s] %(message)s"
VERBOSE_LOG_FORMAT = "%(asctime)s [%(levelname)-8s] {%(process)5d} %(name)s:%(lineno)4d %(message)s"


###########
# Following hack enables 'VERBOSE' log level in the python logging module and Logger class.
# This means you need to import kobo.log before you can use 'VERBOSE' logging.


logging.VERBOSE = 15
logging.addLevelName(15, "VERBOSE")


def verbose(self, msg, *args, **kwargs):
    """
    Log 'msg % args' with severity 'VERBOSE'.

    To pass exception information, use the keyword argument exc_info with
    a true value, e.g.

    logger.info("Houston, we have a %s", "interesting problem", exc_info=1)
    """
    if self.manager.disable >= logging.VERBOSE:
        return
    if logging.VERBOSE >= self.getEffectiveLevel():
        self._log(*(logging.VERBOSE, msg, args), **kwargs)
logging.Logger.verbose = verbose


def verbose(msg, *args, **kwargs):
    """
    Log a message with severity 'VERBOSE' on the root logger.
    """
    if len(logging.root.handlers) == 0:
        logging.basicConfig()
    logging.root.verbose(*((msg, ) + args), **kwargs)
logging.verbose = verbose
del verbose


# end of hack
###########


def add_stderr_logger(logger, log_level=None, format=None):
    """Add a stderr logger to the logger.

    Deprecated, use ``add_stream_handler`` instead.
    """
    add_stream_handler(logger, log_level=log_level, format=format)


def add_stream_handler(logger, log_level=None, format=None, stream=None):
    """Add a handler that prints values to the given stream. Defaults to
    sys.stderr.
    """
    log_level = log_level or logging.DEBUG
    format = format or BRIEF_LOG_FORMAT

    handler = logging.StreamHandler(stream)
    handler.setFormatter(
        logging.Formatter(format, datefmt="%Y-%m-%d %H:%M:%S"))
    handler.setLevel(log_level)
    logger.addHandler(handler)


class KoboLogWarning(UserWarning):
    """Just like a UserWarning, but with a custom name for better filtering."""
    pass


def _warn(msg, *args):
    """Report a warning pointing to a line that called whatever function called
    _warn.
    """
    warnings.warn(msg % args, KoboLogWarning, stacklevel=3)


def add_file_logger(logger, logfile, log_level=None, format=None, mode="a"):
    """Add a file logger to the logger.

    When there is a problem with the log file, a warning is logged. It can be
    turned into an exception by running:

    > warnings.simplefilter('error', kobo.log.KoboLogWarning)
    """
    log_level = log_level or logging.DEBUG
    format = format or BRIEF_LOG_FORMAT

    # Create parent directory if needed.
    try:
        os.makedirs(os.path.dirname(logfile))
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            _warn('Could not create %s: %s', os.path.dirname(logfile), exc)
            return

    # touch the logfile
    if not os.path.exists(logfile):
        try:
            fo = open(logfile, "w")
            fo.close()
        except IOError as exc:
            _warn('Could not touch %s: %s', logfile, exc)
            return

    # is the logfile really a file?
    if not os.path.isfile(logfile):
        _warn('Can not log into %s: not a file', logfile)
        return

    # check if the logfile is writable
    if not os.access(logfile, os.W_OK):
        _warn('Can not log into %s: not writable', logfile)
        return

    handler = logging.FileHandler(logfile, mode=mode)
    handler.setFormatter(logging.Formatter(format, datefmt="%Y-%m-%d %H:%M:%S"))
    handler.setLevel(log_level)
    logger.addHandler(handler)


def add_rotating_file_logger(logger, logfile, log_level=None, format=None, mode="a"):
    """Add a rotating file logger to the logger."""
    log_level = log_level or logging.DEBUG
    format = format or BRIEF_LOG_FORMAT

    # touch the logfile
    if not os.path.exists(logfile):
        try:
            fo = open(logfile, "w")
            fo.close()
        except (ValueError, IOError):
            return

    # is the logfile really a file?
    if not os.path.isfile(logfile):
        return

    # check if the logfile is writable
    if not os.access(logfile, os.W_OK):
        return

    handler = logging.handlers.RotatingFileHandler(logfile, maxBytes=10*(1024**2), backupCount=5, mode=mode)
    handler.setFormatter(logging.Formatter(format, datefmt="%Y-%m-%d %H:%M:%S"))
    handler.setLevel(log_level)
    logger.addHandler(handler)


class LoggingBase(object):
    """Inherit from this class to obtain log_* logging methods."""

    def __init__(self, logger=None):
        self._logger = logger

    def __log(self, level, msg, *args, **kwargs):
        if self._logger is None:
            return
        self._logger.log(level, msg, *args, **kwargs)

    def log_debug(self, msg, *args, **kwargs):
        self.__log(logging.DEBUG, msg, *args, **kwargs)

    def log_verbose(self, msg, *args, **kwargs):
        self.__log(logging.VERBOSE, msg, *args, **kwargs)

    def log_info(self, msg, *args, **kwargs):
        self.__log(logging.INFO, msg, *args, **kwargs)

    def log_warning(self, msg, *args, **kwargs):
        self.__log(logging.WARNING, msg, *args, **kwargs)

    def log_error(self, msg, *args, **kwargs):
        self.__log(logging.ERROR, msg, *args, **kwargs)

    def log_critical(self, msg, *args, **kwargs):
        self.__log(logging.CRITICAL, msg, *args, **kwargs)



from django.contrib.auth.middleware import RemoteUserMiddleware
from kobo.django.django_version import django_version_ge
if django_version_ge('1.10.0'):
    from django.utils.deprecation import MiddlewareMixin

from kobo.django.helpers import call_if_callable

class LimitedRemoteUserMiddleware(RemoteUserMiddleware, MiddlewareMixin if django_version_ge('1.10.0') else object):
    '''
    Same behaviour as RemoteUserMiddleware except that it doesn't logout user
    if is already logged in.
    Useful when you have just one authentication powered login page.
    '''
    def process_request(self, request):
        if not hasattr(request, 'user') or not call_if_callable(request.user.is_authenticated):
            super(LimitedRemoteUserMiddleware, self).process_request(request)
